from ..worker import Worker
from ...mesh.datatypes import *
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ...attributes.misc_faces import face_area

from ...geometry import Vec
from ... import geometry as geom
from ... import utils

import numpy as np

class ParamDistortion(Worker):
    """
    Utility class to compute various distortion metrics for surface parametrization.
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
            mesh : SurfaceMesh,
            uv_attr : str = "uv_coords",
            save_on_mesh : bool = True,
            verbose : bool = False):
        """
        Initializes the distortion utility class.

        Args:
            mesh (SurfaceMesh): the supporting mesh
            uv_attr (str, optional): the attribute name that stores the uv-coordinates on face corners. Defaults to "uv_coords".
            save_on_mesh (bool, optional): If True, distortion values will be stored as attributes on the mesh. Defaults to True.
            verbose (bool, optional): verbose mode. Defaults to False.

        Raises:
            Exception: fails if the attribute 'uv_attr' does not exists.
        """
        super().__init__("SurfaceDistortion", verbose)
        self.mesh : SurfaceMesh = mesh
        try:
            self.UV = self.mesh.face_corners.get_attribute(uv_attr)
        except:
            self.log(f"Mesh has no attribute '{uv_attr}'. Cannot compute distortion.")
            raise Exception("Initialization failed")

        self.save_on_mesh : bool = save_on_mesh

        self._summary : dict = None

        self._conformal : ArrayAttribute = None # ||J||² / det J
        self._scale     : ArrayAttribute = None # 0.5*(det J + 1 / det J)
        self._iso       : ArrayAttribute = None # distance( (sigma_1, sigma_2), (1,1))
        self._shear     : ArrayAttribute = None # dot(c_1, c_2) of columns of jacobian 
        self._stretch   : ArrayAttribute = None # sigma_1 / sigma_2
        self._det       : ArrayAttribute = None # determinant of faces

    def run(self):
        """
        Run the distortion computation.

        Raises:
            Exception: fails if the mesh is not triangular.
            ZeroDivisionError: if degenerated elements are present in the parametrization.
        """
        if not ({len(f) for f in self.mesh.faces} == {3}) :
            # do not call mesh.is_triangular() since we would like to also be compatible with some RawMeshData
            raise Exception("Mesh is not triangular")
        self._init_containers()
        xy_area = 0.
        uv_area = 0.
        area = face_area(self.mesh, persistent=False)
        for T in self.mesh.id_faces:
            cnr = 3*T # self.mesh.connectivity.face_to_first_corner(T)
            xy_area += area[T]
            uvA,uvB,uvC = ( Vec( self.UV[cnr + i][0], self.UV[cnr + i][1], 0.) for i in range(3))
            uv_area += geom.triangle_area(uvA,uvB,uvC)

        scale_ratio = (xy_area / uv_area)

        conformalDist = 0. # ||J||² / det J
        authalicDist = 0. # det J + 1 / det J
        isoDist = 0. # distance( (sigma_1, sigma_2), (1,1))
        shearDist = 0. # dot(c_1, c_2) of columns of jacobian 
        stretchDistMean = 0 # sigma_1 / sigma_2 
        stretchDistMax = -float("inf") # sigma_1 / sigma_2

        for T in self.mesh.id_faces:
            try:
                A,B,C = self.mesh.faces[T]
                cnr = 3*T #self.mesh.connectivity.face_to_first_corner(T)
                pA,pB,pC = (self.mesh.vertices[x] for x in (A,B,C))
                X,Y,N = geom.face_basis(pA,pB,pC)

                # original coordinates of the triangle
                u0 = pB-pA
                v0 = pC-pA
                u0 = complex(X.dot(u0), Y.dot(u0))
                v0 = complex(X.dot(v0), Y.dot(v0))

                # new coordinates of the triangle
                qA,qB,qC = (self.UV[cnr + i] for i in range(3))
                u = qB - qA
                v = qC - qA
                
                # jacobian
                J0 = np.array([[u0.real, v0.real], 
                            [u0.imag, v0.imag]])
                J0 = np.linalg.inv(J0)
                J1 = np.array([[u[0], v[0]], 
                               [u[1], v[1]]])
                J = J1 @ J0

                try:
                    c1, c2 = Vec.normalized(J[:,0]), Vec.normalized(J[:,1])
                    shearDistT = abs(geom.dot(c1, c2))
                except:
                    shearDistT = 0.
                self._shear[T] = shearDistT
                shearDist += shearDistT * area[T] / xy_area

                sig = np.linalg.svd(J, compute_uv=False) # eigenvalues
                detJ = np.linalg.det(J)
                
                if abs(detJ)<1e-8:
                    raise ZeroDivisionError

                confDistT = (np.trace(np.transpose(J) * J)/detJ)/2
                self._conformal[T] = confDistT
                conformalDist += confDistT * area[T] / xy_area

                detJ *= scale_ratio
                self._det[T] = detJ
                
                authDistT = ( detJ + 1 / detJ)/2
                self._scale[T] = authDistT
                authalicDist += authDistT * area[T]/xy_area

                stretchDistT = sig[0]/sig[1]
                self._stretch[T] = stretchDistT
                stretchDistMax = max(stretchDistMax, stretchDistT)
                stretchDistMean += stretchDistT * area[T] / xy_area

                isoDistT = geom.distance( Vec(sig[0]* np.sqrt(scale_ratio), sig[1]* np.sqrt(scale_ratio)), Vec(1.,1.))
                self._iso[T] = isoDistT
                isoDist += isoDistT * area[T] / xy_area

            except ZeroDivisionError:
                continue

        self._summary = {
            "conformal" : conformalDist,
            "iso" : isoDist,
            "shear" : shearDist,
            "scale" : authalicDist,
            "stretch_mean" : stretchDistMean,
            "stretch_max" : stretchDistMax,
        }

    def _init_containers(self):
        if self.save_on_mesh:
            self._conformal = self.mesh.faces.create_attribute("conformal_dist", float, dense=True)
            self._scale = self.mesh.faces.create_attribute("scale_dist", float, dense=True)
            self._stretch = self.mesh.faces.create_attribute("stretch_dist", float, dense=True)
            self._shear = self.mesh.faces.create_attribute("shear_dist", float, dense=True)
            self._iso = self.mesh.faces.create_attribute("iso_dist", float, dense=True)
            self._det = self.mesh.faces.create_attribute("det", float, dense=True)
        else:
            N = len(self.mesh.faces)
            self._conformal = ArrayAttribute(float, N)
            self._scale = ArrayAttribute(float, N)
            self._stretch = ArrayAttribute(float, N)
            self._shear = ArrayAttribute(float, N)
            self._iso = ArrayAttribute(float, N)
            self._det = ArrayAttribute(float, N)

    @property
    def summary(self) -> dict:
        """
        Computes a summary dictionnary of all distortion values as an average over the mesh

        Returns:
            dict: a dictionnary with aggregated values over the mesh
        """
        if self._summary is None: self.run()
        return self._summary

    @property
    def det(self) -> ArrayAttribute:
        """
        Jacobian determinants of each element
        """
        if self._det is None: self.run()
        return self._det

    @property
    def conformal(self) -> ArrayAttribute:
        """
        Conformal distortion, defined as $\\frac{||J||^2}{\\det(J)}$
        """
        if self._conformal is None: self.run()
        return self._conformal
    
    @property
    def scale(self) -> ArrayAttribute:
        """
        Scale distortion, defined as $\\frac12 (\\det(J) + 1/\\det(J))$
        """
        if self._scale is None: self.run()
        return self._scale
    
    @property
    def stretch(self) -> ArrayAttribute:
        """
        Stretch distortion

        Defined as the ratio $\\frac{\\sigma_1}{\\sigma_2}$ where $\\sigma_1$ and $\\sigma_2$ are the eigenvalues of J
        """
        if self._stretch is None: self.run()
        return self._stretch

    @property
    def iso(self) -> ArrayAttribute :
        """
        Isometric distortion

        Defined as the distance from ($\\sigma_1$, $\\sigma_2$) to (1,1) where $\\sigma_1$ and $\\sigma_2$ are the eigenvalues of J
        """
        if self._iso is None: self.run()
        return self._iso

    @property
    def shear(self) -> ArrayAttribute :
        """
        Shear distortion

        Defined as $c1 \\cdot c2$ where $c1$ and $c2$ are the columns of J
        """
        if self._shear is None: self.run()
        return self._shear


class QuadQuality(Worker):
    """
    Utility class to compute various quality metrics on quad meshes
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
            mesh : SurfaceMesh,
            save_on_mesh : bool = True,
            verbose : bool = False):
        super().__init__("SurfaceDistortion", verbose)
        self.mesh : SurfaceMesh = mesh
        self.save_on_mesh = save_on_mesh

        self._summary : dict = None
        self._conformal : ArrayAttribute = None
        self._scale : ArrayAttribute = None
        self._stretch : ArrayAttribute = None
        self._det : ArrayAttribute = None

    def run(self):
        """Runs the distortion computation

        Raises:
            Exception: fails if the mesh is not a quadmesh
        """
        if not ({len(f) for f in self.mesh.faces} == {4}) :
            raise Exception("Mesh is not quad")

        self._init_containers()
        ref_area = len(self.mesh.faces)
        real_area = 0.
        area = face_area(self.mesh, persistent=False)
        for T in self.mesh.id_faces:
            real_area += area[T]

        scale_ratio = (ref_area / real_area)

        conformalDist = 0. # ||J||² / det J
        scaleDist = 0. # 0.5*(det J + 1 / det J)
        detDist = 0. # det J
        stretchDistMean = 0

        for iT,T in enumerate(self.mesh.faces):
            try:
                for v1,v2,v3 in utils.cyclic_triplets(T):
                    P1,P2,P3 = (self.mesh.vertices[_v] for _v in (v1,v2,v3))
                    X,Y,_ = geom.face_basis(P1,P2,P3)
                    # new coordinates of the triangle
                    u = P1 - P2
                    v = P3 - P2
                    # jacobian
                    J = np.array([[X.dot(u), Y.dot(u)], 
                                [X.dot(v), Y.dot(v)]])

                    sig = np.linalg.svd(J, compute_uv=False)
                    detJ = np.linalg.det(J)
                    
                    if abs(detJ)<1e-8:
                        raise ZeroDivisionError

                    normJ = np.linalg.norm(J)

                    confDistT = ((normJ*normJ)/detJ)/8
                    self._conformal[iT] += confDistT
                    conformalDist += confDistT * area[iT] / real_area

                    detJ *= scale_ratio
                    detDist += detJ / (4*len(self.mesh.faces))
                    self._det[iT] += detJ/4
                    scaleDistT = ( detJ + 1 / detJ)/8

                    self._scale[iT] += scaleDistT
                    scaleDist += scaleDistT * area[iT] / real_area

                    stretchDistT = (sig[0]/sig[1])/4
                    self._stretch[iT] += stretchDistT
                    stretchDistMean += stretchDistT * area[iT] / real_area
            except ZeroDivisionError:
                continue
            except FloatingPointError:
                continue

        self._summary = {
            "conformal" : conformalDist,
            "scale" : scaleDist,
            "det" : detDist,
            "stretch" : stretchDistMean
        }
        

    def _init_containers(self):
        if self.save_on_mesh:
            self._conformal = self.mesh.faces.create_attribute("conformal_dist", float, dense=True)
            self._scale = self.mesh.faces.create_attribute("scale_dist", float, dense=True)
            self._stretch = self.mesh.faces.create_attribute("stretch_dist", float, dense=True)
            self._det = self.mesh.faces.create_attribute("det", float, dense=True)
        else:
            N = len(self.mesh.faces)
            self._conformal = ArrayAttribute(float, N)
            self._scale = ArrayAttribute(float, N)
            self._stretch = ArrayAttribute(float, N)
            self._det = ArrayAttribute(float, N)

    @property
    def summary(self):
        if self._summary is None: self.run()
        return self._summary

    @property
    def conformal(self):
        if self._conformal is None: self.run()
        return self._conformal
    
    @property
    def area(self):
        if self._scale is None: self.run()
        return self._scale
    
    @property
    def stretch(self):
        if self._stretch is None: self.run()
        return self._stretch

    @property
    def det(self):
        if self._det is None: self.run()
        return self._det

    @property
    def iso(self):
        if self._iso is None: self.run()
        return self._iso

    @property
    def shear(self):
        if self._shear is None: self.run()
        return self._shear


import scipy.sparse as sp
import numpy as np
from .. import geometry as geom

def inverse_power_method(A: sp.csc_matrix, m: float = 0., B: sp.csc_matrix = None, maxiter: int = 100, tol: float=1e-10) -> np.ndarray:
    """
    Implementation of the inverse power method (or inverse iteration) scheme to find an eigenvector associated with an eigenvalue of A that is close to 'm'.
    In other words, this function computes x such that : Ax = λBx where λ is an eigenvalue of A that minimizes |λ-m|.

    Args:
        A (sp.csc_matrix): the matrix
        mu (float): the approximate eigenvalue. Will compute an eigenvector for the eigenvalue λ that minimizes its distance to mu. Defaults to zero.
        B (sp.csc_matrix, optional): metrics matrix for generalized eigenvectors computation. If not specified, will be the identity matrix. Defaults to None.
        maxiter (int, optional): maximal number of internal iteration. Defaults to 100.
        tol (float, optional): early stopping criterion. Will stop the iteration if |x_{n+1} - x_n| < tol. Defaults to 1e-6.

    Returns:
        np.ndarray: a unit norm eiven vector associated with an eigenvalue that is close to 'mu'
    """
    n = A.shape[0]
    B = sp.eye(n, format="csc") if B is None else B
    solve = sp.linalg.factorized(A - m * sp.eye(n))
    x = np.random.random(n)

    A_is_hermitian = (A.dtype==complex)
    if A_is_hermitian:
        x = np.sqrt(np.random.uniform(0, 1, n)) * np.exp(1.j * np.random.uniform(0, 2 * np.pi, n)) # values in unit disk in complex plane
    else:
        x = 2*np.random.random(n)-1 # values in [-1,1]

    it = 0
    stop_criterion = False
    while not stop_criterion:
        newx = solve(B@x)
        newxT = newx.conj() if A_is_hermitian else newx
        newx /= np.sqrt(np.dot(newxT, B@newx))
        it += 1
        stop_criterion = (it>=maxiter or geom.distance(newx, x)<tol)
        x = newx
    return x


def rayleigh_quotient_iteration(
        A: sp.csc_matrix, 
        m: float = 0., 
        B: sp.csc_matrix = None, 
        x0: np.ndarray = None, 
        maxiter: int = 100, 
        tol: float = 1e-10
    ) -> np.ndarray:
    
    n = A.shape[0]
    B = sp.eye(n, format="csc") if B is None else B
    
    if x0 is not None:
        x = np.copy(x0)
    else:
        A_is_hermitian = (A.dtype==complex)
        if A_is_hermitian:
            x = np.sqrt(np.random.uniform(0, 1, n)) * np.exp(1.j * np.random.uniform(0, 2 * np.pi, n)) # values in unit disk in complex plane
        else:
            x = 2*np.random.random(n)-1 # values in [-1,1]
        x /= np.linalg.norm(x)

    for it in range(maxiter):
        print(it)
        y = sp.linalg.spsolve(A - m*B, B@x)
        m = (y.T @ A @ y) / (y.T @ B @ y)
        y /= np.linalg.norm(y)
        if np.linalg.norm(x-y) < tol: break # converged
        x = y
    return m,x


from numba import njit, prange, set_num_threads
import numpy as np
from scipy.optimize import fmin_l_bfgs_b
from .base import BaseParametrization
from ...mesh.datatypes import *
from ...mesh.mesh import copy
from ...mesh.mesh_attributes import ArrayAttribute
from ... import geometry as geom
from ...attributes.glob import euler_characteristic
from ...attributes.misc_faces import face_area

@njit(cache=True, fastmath=True)
def chi(D, eps):
    # regularization function
    return (D + np.sqrt(eps*eps + D*D))/2

@njit(cache=True, fastmath=True)
def chi_deriv(D,eps):
    return 0.5 + D/(2 * np.sqrt(eps*eps + D*D))

@njit(cache=True, fastmath=True)
def jacobian(pts, tri, ref_jac):
    iA,iB,iC = tri[0], tri[1], tri[2]
    A,B,C = pts[2*iA:2*iA+2], pts[2*iB:2*iB+2], pts[2*iC:2*iC+2]
    return np.vstack((B-A, C-A)).T @ ref_jac

@njit(cache=True, parallel=True)
def min_det(pts, tris, ref_jacs):
    n_tri = tris.shape[0]
    dets = np.zeros(n_tri, dtype=np.float64)
    for t in prange(n_tri):
        dets[t] = np.linalg.det(jacobian(pts, tris[t], ref_jacs[t]))
    return np.min(dets)

@njit(cache=True, parallel=True)
def winslow_energy_and_gradient(X, locked, tri, ref_jacs, areas, wf, wg, eps):
    E = 0
    grad = np.zeros_like(X)
    n = tri.shape[0]
    Z = np.array([[-1.,-1.], [1.,0.], [0.,1.]])
    for t in prange(n):
        J = jacobian(X, tri[t], ref_jacs[t])
        K = np.array([[J[1,1],-J[1,0]],[-J[0,1],J[0,0]]])
        detJ = np.linalg.det(J)
        chiJ = chi(detJ, eps)
        fJ = np.trace( np.transpose(J) @ J) / chiJ
        gJ = (detJ*detJ + 1) / chiJ
        # hJ = chiJ - np.log(chiJ)
        
        E += areas[t] * (wf * fJ  + wg * gJ)
        # E += areas[t] * ( wf * fJ + wg * hJ )

        # compute gradient according to J
        chi_derivJ = chi_deriv(detJ, eps)
        ZJ = Z @ ref_jacs[t]
        
        df_dj = (2*J -  K * fJ * chi_derivJ) / chiJ
        dg_dj = (2*detJ - gJ * chi_derivJ) / chiJ * K
        # dh_dj = (chi_derivJ * (1 - 1/chiJ)) * K
        dphi_dj = areas[t] * (wf * df_dj + wg * dg_dj)
        # dphi_dj = areas[t] * (wf * df_dj + wg * dh_dj)
        dphi_du = ZJ @ dphi_dj.T # chain rule for the actual variables
        for i,v in enumerate(tri[t]):
            if locked[v] : continue
            grad[2*v  ] += dphi_du[i,0]
            grad[2*v+1] += dphi_du[i,1]
    return E,grad

def untangle(
    points: np.ndarray,
    locked: np.ndarray,
    triangles: np.ndarray,
    ref_jacs: np.ndarray,
    verbose: bool = False,
    **kwargs
) -> np.ndarray :
    """
    Minimizes the regularized Winslow functional to untangle a 2D triangulation.

    Args:
        points (np.ndarray[float]): Initial position of points in 2D. Should be of shape (2*V,)
        locked (np.ndarray[bool]): Which points have a fixed position. Should be of shape (V,)
        triangles (np.ndarray[int]): Indices of triangles. Should be of shape (T,3)
        ref_jacs (np.ndarray[float]): Perfect element to consider for Jacobian computation for each triangle. Should be of shape (T,2,2).
        verbose (bool,optional): Verbose mode. Defaults to False.
    
    Keyword Args:
        areas (np.ndarray[float]): Areas of triangles in original mesh. Used as a weighting term in the energy's summation. Should be of shape (T,). Defaults to np.ones(T).
        weight_angles (float, optional): weight coefficient for the angle conservation term (f). Defaults to 1.
        weight_areas (float, optional): weight coefficient for the area conservation term (g). Defaults to 1.
        iter_max (int, optional): Maximum number of iterations in the L-BFGS solver. Defaults to 10000.
        n_eps_update (int, optional): number of updates of the regularization's epsilon. Defaults to 10.
        stop_if_positive (bool, optional): enable early stopping as soon as all dets are positive. Defaults to False.

    Returns:
        np.ndarray[float]: final positions of points in 2D, in shape (2V,)

    References:
        [1] _Foldover-free maps in 50 lines of code_, Garanzha et al., ACM ToG 2021
    """
    n_eps_update = kwargs.get("n_eps_update", 10)
    wf = kwargs.get("weight_angles", 1.)
    wg = kwargs.get("weight_areas", 1.)
    bfgs_iter_max = kwargs.get("iter_max", 10_000)
    stop_if_pos = kwargs.get("stop_if_positive",False)
    areas = kwargs.get("areas", None)
    set_num_threads(2)
        
    if areas is None:
        areas = np.ones(triangles.shape[0])

    if verbose:
        print("WEIGHT ANGLES:", wf)
        print("WEIGHT AREAS:", wg)

    for _ in range(n_eps_update):
        mindet = min_det(points, triangles, ref_jacs)
        if mindet>0 and stop_if_pos: break
        eps = np.sqrt(1e-12 + .04*min(mindet, 0)**2) # the regularization parameter e
        if verbose:
            print("min det=", mindet)
            print("epsilon=", eps)
        points = fmin_l_bfgs_b(
            winslow_energy_and_gradient, points, 
            args=(locked, triangles, ref_jacs, areas, wf, wg, eps),
            disp=10 if verbose else 0, maxiter=bfgs_iter_max
        )[0]
    
    if verbose: print("min det=", min_det(points, triangles, ref_jacs))

    return points


class WinslowInjectiveEmbedding(BaseParametrization):
    """
    Foldover-free map to the plane: computes an injective embedding in the plane starting from the provided $uv$-coordinates by minimizing the regularized Winslow functionnal
    of Garanzha et al. [1]. This class is essentially a wrapper around the `untangle` function.
    
    Warning: 
        The input mesh should have the topology of a disk.

        UV coordinates are computed per vertex and not per corner. See `mouette.attributes.scatter_vertices_to_corners` for conversion.

    References:
        [1] _Foldover-free maps in 50 lines of code_, Garanzha et al., ACM ToG 2021

    Example:
        See [https://github.com/GCoiffier/mouette/blob/main/examples/winslow_untangle.py](https://github.com/GCoiffier/mouette/blob/main/examples/winslow_untangle.py)
    """
            
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, uv_init: ArrayAttribute, verbose : bool=True, **kwargs):
        """
        Args:
            mesh (SurfaceMesh): the supporting mesh. Should be a surface with disk topology.
            uv_init (ArrayAttribute): array of initial uv-coordinates per vertices. np.array of shape (V,2) or mouette.ArrayAttribute object.
            verbose (bool, optional): verbose mode. Defaults to True.
        
        Keyword Args:
            stop_if_positive (bool, optional): whether to stop the optimization as soon as all determinants are positive. Defaults to False.
            solver_verbose (bool, optional): Verbose tag for the L-BFGS solver. Defaults to False.

        Attributes:
            uvs (Attribute): an attribute containing the uv-coordinates of the parametrization
        """
        super().__init__("Winslow", mesh, verbose=verbose, uv_attr=uv_init, **kwargs)
        self._save_on_corners = False
        self._solver_verbose = kwargs.get("solver_verbose", False)
        self.lmbd = kwargs.get("lmbd", 1.)
        self.stop_if_pos = kwargs.get("stop_if_positive", False)
        self.ref_jacs = kwargs.get("ref_jacs", None)
        
    def run(self):
        """
        Calls the solver.

        Raises:
            Exception: fails if the mesh is not a triangulation of a topological disk
        """
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk")
        if not self.mesh.is_triangular():
            raise Exception("Mesh is not triangular.")

        if self.ref_jacs is not None:
            ref_jacs = np.asarray(self.ref_jacs)
        else:
            ref_jacs = np.array([[[1.,0.], [0.,1.]] for _ in self.mesh.id_faces]) # reference element
        source_area = sum([geom.triangle_area_2D(self.uvs[A], self.uvs[B], self.uvs[C]) for (A,B,C) in self.mesh.faces])        
        scale = np.sqrt(len(self.mesh.faces)/source_area)

        points = np.concatenate([self.uvs[v] for v in self.mesh.id_vertices]) 
        points *= scale
        locked = np.array([self.mesh.is_vertex_on_border(v) for v in self.mesh.id_vertices])
        triangles = np.array(self.mesh.faces)
        
        areas = face_area(self.mesh, persistent=False, dense=True).as_array()

        final_points = untangle(points, locked, triangles, ref_jacs,
            areas=areas, weight_angles=1., weight_areas=self.lmbd, 
            verbose=self.verbose, stop_if_positive = self.stop_if_pos)
        final_points /= scale

        # Retrieve uvs and write them in attribute
        self.uvs = self.mesh.vertices.create_attribute("uv_coords", float, 2, dense=True)
        for v in self.mesh.id_vertices:
            self.uvs[v] = final_points[2*v:2*v+2]

    @property
    def flat_mesh(self) -> SurfaceMesh:
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None: return None
        flat = copy(self.mesh)
        # build the flat mesh : vertex coordinates are uv of original mesh
        for T in self.mesh.id_faces:
            for i,v in enumerate(self.mesh.faces[T]):
                flat.vertices[v] = geom.Vec(self.uvs[v][0], self.uvs[v][1], 0.)
        return flat


from ...utils import get_extension
from .geogram_ascii import import_geogram_ascii, export_geogram_ascii
from .obj import import_obj, export_obj
from .medit import import_medit, export_medit
from .off import import_off, export_off
from .xyz import import_xyz, export_xyz
from .tet import import_tet, export_tet
from .stl import import_stl, export_stl
from .ply import import_ply, export_ply

def read_by_extension(filename : str):
    ext = get_extension(filename)
    import_fun = {
        "geogram_ascii" : import_geogram_ascii,
        "obj" : import_obj,
        "mesh" : import_medit,
        "off" : import_off,
        "xyz" : import_xyz,
        "tet" : import_tet,
        "stl" : import_stl,
        "ply" : import_ply
    }.get(ext.lower(), None)
    if import_fun is None:
        raise Exception("Unsupported file extension '{}'".format(ext))
    data = import_fun(filename)
    return data

def write_by_extension(mesh : "RawMeshData", filename : str):
    ext = get_extension(filename)
    export_fun = {
        "geogram_ascii" : export_geogram_ascii,
        "obj" : export_obj,
        "mesh" : export_medit,
        "off" : export_off,
        "xyz" : export_xyz,
        "tet" : export_tet,
        "stl" : export_stl,
        "ply" : export_ply
    }.get(ext.lower(), None)
    if export_fun is None:
        raise Exception("Unsupported file extension '{}'".format(ext))
    export_fun(mesh, filename)


import heapq as hq
from typing import Any
from dataclasses import dataclass, field

@dataclass
class PriorityItem:
    """
    Dataclass representing an item inside the priority queue.
    An item is a pair element/priority.
    """
    x : Any = field(compare=False)
    priority: float

    def __lt__(self, other):
        return self.priority < other.priority

class PriorityQueue:
    """
    Implementation of a priority queue using a heap from the heapq package.

    Attributes:
        data (list): the list representing the queue
    """

    def __init__(self):
        """
        Initializes an empty queue.
        """
        self.data = []

    def empty(self) -> bool:
        """
        Tests if the queue contains any element.

        Returns:
            bool: True if the queue is empty
        """
        return len(self.data) == 0

    @property
    def front(self) -> PriorityItem:
        """
        The first element of the queue. Is accessed but not removed from the queue

        Returns:
            PriorityItem: The first element of the queue
        """
        return self.data[0]

    def get(self) -> PriorityItem:
        """
        Returns the first element of the queue. The element is removed from the queue.

        Returns:
            PriorityItem: The first element of the queue
        
        Raises:
            IndexError: if the queue is empty
        """
        return hq.heappop(self.data)

    def pop(self) -> PriorityItem:
        """
        Returns the first element of the queue. Same method as `get`.

        Returns:
            PriorityItem: The first element of the queue

        Raises:
            IndexError: if the queue is empty
        """
        return self.get()
    
    def push(self, x, w : float) -> None:
        """
        Inserts an element inside the queue.

        Args:
            x: Element to insert in the queue
            w: Priority of the element.
        """
        item = PriorityItem(x, w)
        hq.heappush(self.data, item)



"""
argument_check.py

Utility functions and exceptions for sanity check on function arguments
"""

class InvalidArgumentTypeError(Exception):
    def __init__(self, argname :str, argtype : type, *expected_types:type):
        message = f"Argument '{argname}' has type {argtype} while the function expected types {expected_types}"
        super().__init__(message)

class InvalidArgumentValueError(Exception):
    def __init__(self, argname :str, argvalue : type, expected_values:list = None):
        if expected_values is None:
            message = f"Argument '{argname}' has invalid value '{argvalue}'"
        else:
            message = f"Argument '{argname}' has invalid value {argvalue}. Allowed values are {expected_values}"
        super().__init__(message)

class InvalidRangeArgumentError(Exception):
    def __init__(self, argname:str, argvalue, condition:str):
        message = f"Argument {argname} is out of range. Expected a value {condition}. Got {argvalue}"
        super().__init__(message)

class NegativeArgumentError(Exception):
    def __init__(self, argname:str, argvalue):
        message = f"Argument '{argname}' has value {argvalue} but should be >=0"
        super().__init__(message)

def check_argument(argname :str, argvalue, expected_type: type, expected_values:list=None):
    """
    check input sanity on the argument of a function

    Args:
        argname (str): Name of the argument (for exhaustivity of error message)
        argvalue (any): value of the argument
        
        expected_type (type): expected type of the argument. 
            If type(argvalue) does not match, will raise an InvalidArgumentTypeError

        expected_values (list, optional): List of possible values the argument is allowed to take. 
            If argvalue is not in the list, will raise an InvalidArgumentValueError.
            Defaults to None.
    """
    if not isinstance(argvalue, expected_type):
        raise InvalidArgumentTypeError(argname, type(argvalue), expected_type)
    if expected_values is not None and argvalue not in expected_values:
        raise InvalidArgumentValueError(argname, argvalue, expected_values)


import stl_reader
import struct
from ..mesh_data import RawMeshData
from collections import deque
import warnings

def is_stl_ascii(path : str):
    try:
        with open(path, 'r', encoding="utf-8") as f:
            return "solid" in f.readline().strip().split()
    except Exception as e:
        return False

def import_stl(path : str):
    if is_stl_ascii(path):
        return _import_stl_ascii(path)
    else:
        vertices, faces =  stl_reader.read(path)
        out = RawMeshData()
        out.vertices += list(vertices)
        out.faces += list(faces)
        return out

def _import_stl_ascii(path : str):
    data = deque()
    with open(path, "r", encoding="utf-8") as stlf:
        data = deque([x.strip().split() for x in stlf.readlines()])
    
    out = RawMeshData()
    normals = out.faces.create_attribute("normals", float, 3)
    iF = 0
    while data:
        line = data.popleft()
        if line[0]=="solid": continue
        
        elif line[0]=="facet":
            assert line[1] == "normal"
            normals[iF] = [float(x) for x in line[2:]]
            out_loop = data.popleft()
            assert out_loop[0] == "outer"
            vertices = []
            for _ in range(3):
                v = data.popleft()
                assert v[0] == "vertex"
                vertices.append([float(x) for x in v[1:]])
            end_loop = data.popleft()
            assert end_loop[0] == "endloop"
            endfacet = data.popleft()
            assert endfacet[0] == "endfacet"

            out.vertices += vertices
            out.faces.append((3*iF,3*iF+1,3*iF+2))
            iF += 1

        elif line[0] == "endsolid": continue
    return out

# Exporter adapted from https://gist.github.com/ryansturmer/9329299
# License: MIT License 

def export_stl(mesh, path :str):
    if not hasattr(mesh, "faces"):
        warnings.warn("No faces detected in the object. Nothing to write. Returning.")
        return
    with open(path, 'wb') as fp:
        writer = Binary_STL_Writer(fp)
        writer.write(mesh)

class Binary_STL_Writer:
    BINARY_HEADER ="80sI"
    BINARY_FACET = "12fH"

    def __init__(self, stream):
        self.counter = 0
        self.fp = stream
        self._write_header()

    def _write_header(self):
        self.fp.seek(0)
        self.fp.write(struct.pack(Binary_STL_Writer.BINARY_HEADER, b'Python Binary STL Writer', self.counter))

    def _write_triangle(self, p1, p2, p3):
        self.counter += 1
        data = [
            0., 0., 0.,
            p1[0], p1[1], p1[2],
            p2[0], p2[1], p2[2],
            p3[0], p3[1], p3[2],
            0
        ]
        self.fp.write(struct.pack(Binary_STL_Writer.BINARY_FACET, *data))

    def write(self, mesh):
        self._write_header()
        for face in mesh.faces:
            pts = [mesh.vertices[v] for v in face]
            if len(face) == 3:
                self._write_triangle(*pts)
            elif len(face) == 4:
                self._write_triangle(pts[0], pts[1], pts[2])
                self._write_triangle(pts[2], pts[3], pts[0])
            else:
                raise ValueError('Only triangular and quad faces are supported.')
        self._write_header() # rewrite the header with the correct value of self.counter


from .mesh_data import RawMeshData
from .datatypes import Mesh, PointCloud, PolyLine, SurfaceMesh, VolumeMesh
from .io.io import read_by_extension, write_by_extension

import numpy as np

def _instanciate_raw_mesh_data(mesh_data : RawMeshData, dim : int = None) -> Mesh:
    mesh_data.prepare()
    if dim is None: dim = -1
    dim = max(dim, mesh_data.dimensionality)
    if dim==0: return PointCloud(mesh_data)
    if dim==1: return PolyLine(mesh_data)
    if dim==2: return SurfaceMesh(mesh_data)
    if dim==3: return VolumeMesh(mesh_data)

def load(filename: str, dim: int = None, raw: bool = False) -> Mesh:
    """
    Loads a Mesh object from a file on the disk.

    Supported file formats are:
    
    - wavefront (.obj)
    
    - medit (.mesh)
    
    - geogram (.geogram_ascii)
    
    - .ply
    
    - .stl
    
    - .off
    
    - .tet
    
    - .xyz (point clouds only)

    Parameters:
        filename (str):  the path of the file
        dim (int, optional): 
            Override for the dimensionality of the resulting mesh.
            Depending on this value, returns a PointCloud, LineMesh, SurfaceMesh or VolumeMesh. 
            If not specified, computed from the data read in the file.
        raw (bool, optional): Returns the RawMeshData object instead of a fully prepared Mesh object. Defaults to False

    Returns:
        Mesh:  A new mesh object (PointCloud, LineMesh, SurfaceMesh or VolumeMesh)
    """
    data = read_by_extension(filename)
    if raw: return data
    return _instanciate_raw_mesh_data(data, dim)
    

def save(mesh : Mesh, filename: str, ignore_elements:set = None) -> None:
    """ Saves the mesh data into a file.

    Supported file formats are:
    
    - wavefront (.obj)
    
    - medit (.mesh)
    
    - geogram (.geogram_ascii)
    
    - .ply
    
    - .stl
    
    - .off
    
    - .tet
    
    - .xyz (point clouds only)

    Parameters:
        filename (str): The output file path

    Raises:
        Exception: Unsupported file extension
    """
    if isinstance(mesh,VolumeMesh) and ".geogram" in filename:
        mesh.connectivity._compute_adjacent_cell()
    raw_mesh = RawMeshData(mesh) # get rid of connectivity and additional attributes depending on dimension
    if ignore_elements is not None:
        if "edges" in ignore_elements: raw_mesh.edges.clear()
        if "faces" in ignore_elements:
            raw_mesh.faces.clear()
            raw_mesh.face_corners.clear()
        if "cells" in ignore_elements: 
            raw_mesh.cells.clear()
            raw_mesh.cell_corners.clear()
            raw_mesh.cell_faces.clear()
    write_by_extension(raw_mesh, filename)

def from_arrays(
        V: np.ndarray, 
        E: np.ndarray = None, 
        F: np.ndarray = None, 
        C: np.ndarray = None, 
        raw: bool = False
    ) -> Mesh:
    """Creates a mesh object from numpy arrays.

    Args:
        V (np.ndarray): vertex coordinates. SHould have shape |V|*3 or |V|*2. If shape |V|*2, the array is padded so that z=0.
        E (np.ndarray, optional): Edge indices (shape |E|*2). Defaults to None.
        F (np.ndarray, optional): Face indices (shape |F|*n for n-regular faces). Defaults to None.
        C (np.ndarray, optional): Cell indices (shape |C|*n for n-regular cells). Defaults to None.
        raw (bool, optional): Returns the RawMeshData object instead of a fully prepared Mesh object. Defaults to False

    Returns:
        Mesh: a mesh object (PointCloud to VolumeMesh depending on the data provided)
    """
    m = RawMeshData()
    if V.shape[1]<3:
        V = np.pad(V, ((0,0),(0,3-V.shape[1])))
    elif V.shape[1]!=3: 
        raise Exception("Vertex array should have shape (n,3)")
    
    n_vert = V.shape[0]
    m.vertices += list(V)
    if E is not None:
        if np.any(np.asarray(E)>=n_vert): raise Exception("Edges indices should be between 0 and n_vertices")
        if E.shape[1]!=2: raise Exception("Edge array should have shape (n,2)")
        m.edges += list(E)
    if F is not None:
        if np.any(np.asarray(F)>=n_vert): raise Exception("Face indices should be between 0 and n_vertices")
        m.faces += list(F)
    if C is not None:
        if np.any(np.asarray(C)>=n_vert): raise Exception("Cell indices should be between 0 and n_vertices")
        m.cells += list(C)
    if raw: return m
    return _instanciate_raw_mesh_data(m)

def copy(mesh : Mesh, copy_attributes=False, copy_connectivity=False) -> Mesh:
    """Makes a copy of the input mesh

    Parameters:
        mesh (Mesh): input mesh
        copy_attributes (bool, optional): whether to also copy attributes data. Defaults to False.
        copy_connectivity (bool, optional): whether to also copy the connectivity data. Defaults to False.
    Returns:
        Mesh: a hard copy of the given mesh
    """
    from copy import deepcopy
    copy_mesh = type(mesh)() # instanciate a new mesh of the same type

    if copy_attributes:
        copy_mesh.vertices = deepcopy(mesh.vertices)
        if hasattr(mesh, "edges") : copy_mesh.edges = deepcopy(mesh.edges)
        if hasattr(mesh, "faces") :
            copy_mesh.faces = deepcopy(mesh.faces)
            copy_mesh.face_corners = deepcopy(mesh.face_corners)
        if hasattr(mesh, "cells"):
            copy_mesh.cells = deepcopy(mesh.cells)
            copy_mesh.cell_corners = deepcopy(mesh.cell_corners)
            copy_mesh.cell_faces = deepcopy(mesh.cell_faces)
    else:
        # copy only _cont data of each container (connectivity data and not attribute data)
        copy_mesh.vertices._data = deepcopy(mesh.vertices._data)
        if hasattr(mesh, "edges"): copy_mesh.edges._data = deepcopy(mesh.edges._data)
        if hasattr(mesh, "faces") : 
            copy_mesh.faces._data = deepcopy(mesh.faces._data)
            copy_mesh.face_corners._elem = deepcopy(mesh.face_corners._elem)
            copy_mesh.face_corners._adj = deepcopy(mesh.face_corners._adj)
        if hasattr(mesh, "cells") : 
            copy_mesh.cells._data = deepcopy(mesh.cells._data)
            copy_mesh.cell_corners._elem = deepcopy(mesh.cell_corners._elem)
            copy_mesh.cell_corners._adj = deepcopy(mesh.cell_corners._adj)
            copy_mesh.cell_faces._elem = deepcopy(mesh.cell_faces._elem)
            copy_mesh.cell_faces._adj = deepcopy(mesh.cell_faces._adj)
        # _cont of face_corners and cell_corners are always empty
    if copy_connectivity and hasattr(mesh, "connectivity"):
        copy_mesh.connectivity = mesh.connectivity
    return copy_mesh

def merge(mesh_list : list) -> Mesh:
    """Merges a list of independents meshes as a unique mesh.

    Parameters:
        mesh_list (list): a list of meshes. If empty, this function returns None

    Returns:
        Mesh: the merged mesh. Takes the type of the mesh with the largest dimensionality in the list (ie merging a Polyline with a Volume mesh returns a Volume mesh)
    """
    if len(mesh_list)==0: return None
    merged = RawMeshData()
    vertex_offset = 0
    for to_merge in mesh_list:
        merged.vertices += to_merge.vertices
        if hasattr(to_merge, "edges") : 
            merged.edges += [tuple((vertex_offset+u for u in e)) for e in to_merge.edges]
        if hasattr(to_merge, "faces") : 
            merged.faces += [tuple((vertex_offset+u for u in f)) for f in to_merge.faces]
        if hasattr(to_merge, "cells") :
            merged.cells += [tuple((vertex_offset+u for u in c)) for c in to_merge.cells]
        vertex_offset += len(to_merge.vertices)
    return _instanciate_raw_mesh_data(merged)


def reorder_vertices(mesh : Mesh, new_indices: list) -> Mesh:
    """Reorders vertex indices in a mesh object.

    Args:
        mesh (Mesh): input mesh
        new_indices (list): a list containing a permutation of vertices. New index of vertex L[i] will be i.

    Returns:
        Mesh: the mesh object with permuted vertices
    """
    assert len(new_indices) == len(mesh.vertices)
    ind = np.argsort(new_indices) # "inverted list"
    raw = RawMeshData()
    for v in mesh.id_vertices:
        raw.vertices.append(mesh.vertices[new_indices[v]])
    if hasattr(mesh, "edges"):
        for (A,B) in mesh.edges:
            raw.edges.append((ind[A], ind[B]))
    if hasattr(mesh, "faces"):
        for F in mesh.faces:
            raw.faces.append([ind[v] for v in F])
    if hasattr(mesh, "cells"):
        for C in mesh.cells:
            raw.cells.append([ind[v] for v in C])
    return _instanciate_raw_mesh_data(raw)


from .shapes import tetrahedron, hexahedron, axis_aligned_cube, hexahedron_4pts, octahedron, icosahedron, dodecahedron, cylinder, sphere_uv, sphere_fibonacci, icosphere, torus
from .rings import ring, flat_ring
from .flat import triangle, quad, unit_grid, unit_triangle
from .polylines import chain_of_vertices, vector_field
from .transformations import spherify_vertices, cylindrify_edges
from .dual import dual_mesh


from ..mesh_data import RawMeshData
import pyminiply
import numpy as np

def import_ply(path : str):
    vertices, faces, normals, uv, color = pyminiply.read(path)
    out = RawMeshData()
    out.vertices += list(vertices)
    if np.any(np.asarray(faces)>=len(out.vertices)): raise Exception("Face indices should be between 0 and n_vertices")
    out.faces += list(faces)
    if normals.size>0:
        out.vertices.register_array_as_attribute("normals", normals)
    if uv.size>0:
        out.vertices.register_array_as_attribute("uv_coords", uv)
    if color.size>0:
        out.vertices.register_array_as_attribute("color", color)
    return out


def export_ply(mesh, path : str):
    raise NotImplementedError


from .mesh_attributes import Attribute, ArrayAttribute
from .. import config
import warnings
from abc import ABC, abstractmethod
import numpy as np

class _BaseDataContainer(ABC):
    """
    Base class for a data container. A container encapsulate all elements of the same type in a mesh as well as attributes defined on them.
    """
    def __init__(self, attributes : dict = None, id : str = "") -> None:
        self.id = id
        if attributes is None:
            self._attr = dict()
        else:
            assert isinstance(attributes, dict)
            self._attr = attributes
    
    @abstractmethod
    def empty(self) -> bool:
        pass

    @abstractmethod
    def clear(self):
        pass

    @property
    def attributes(self):
        return self._attr.keys()

    def create_attribute(self, name: str, data_type: type, elem_size: int=1, dense: bool=False, default_value=None, size:int=None) -> Attribute:
        """Creates an attribute with name `name`

        Args:
            name (str): attribute's name. If 'name' already exists in the attribute dict, the corresponding attribute will be overridden. The override warning can be disabled in the config.
            data_type (type): The type of data stored in the attribute.
            elem_size (int, optional): The number of data per elements. Defaults to 1.
            dense (bool, optional): Whether to create an `Attribute` object (dictionnary storage) or an `ArrayAttribute` object (numpy array storage). Defaults to False (dictionnary mode).
            default_value (Any, optional): The default value to fill the attribute with. If not provided, will be taken as the default value of the type `data_type`. Defaults to None.
            size (int, optional): Current size of the DataContainer to be passed. Defaults to None.

        Returns:
            Attribute:
        """
        # If 'name' already exists in the attribute dict, the corresponding attribute will be overridden
        if name in self._attr and config.display_duplicate_attribute_warning:
            warnings.warn(f"Attribute '{name}' already exists on {self.id}")
        else:
            if dense:
                self._attr[name] = ArrayAttribute(data_type, len(self) if size is None else int(size), elem_size=elem_size, default_value=default_value)
            else:
                self._attr[name] = Attribute(data_type, elem_size=elem_size, default_value=default_value)
        return self._attr[name]

    def register_array_as_attribute(self, name: str, data: np.ndarray, default_value=None):
        """Converts a numpy array as an attribute.

        Args:
            name (str): attribute's name. If 'name' already exists in the attribute dict, the corresponding attribute will be overridden. The override warning can be disabled in the config.
            data (np.ndarray): the data array. Type of the attribute will be extrapolated from `data.dtype`.
            default_value (Any, optional): The default value to fill the attribute with. If not provided, will be taken as the default value of the type `data_type`. Defaults to None.
        """
         # If 'name' already exists in the attribute dict, the corresponding attribute will be overridden
        if name in self._attr and not config.display_duplicate_attribute_warning:
            warnings.warn(f"Attribute '{name}' already exists on {self.id}")
        else:
            if len(data.shape)==1: 
                data = data[:,np.newaxis] # change array of size (n,) to size (n,1)
            try:
                n_elem = data.shape[0]
                elem_size = data.shape[1]
                assert n_elem == len(self)
            except Exception as e:
                raise Exception(f"data array has invalid shape {data.shape}")
            self._attr[name] = ArrayAttribute(type(data[0,0].item()), n_elem, elem_size=elem_size, default_value=default_value)
            self._attr[name]._data = data
            return self._attr[name]
        
    def delete_attribute(self, name: str):
        """Deletes the attribute associated with name `name` if it exists.

        Args:
            name (str): the attribute's name.
        """
        if name in self._attr:
            del self._attr[name]

    def has_attribute(self, name: str) -> bool:
        """Returns True is an attribute with name `name` is defined for this container.

        Args:
            name (str): the attribute's name

        Returns:
            bool:
        """
        return name in self._attr

    def get_attribute(self, name: str) -> Attribute :
        """Returns the attribute with name `name`. Fails if the attribute does not exist

        Args:
            name (str): the attribute's name

        Raises:
            Exception: no attribute has name `name` inside this data container.

        Returns:
            Attribute: the attribute
        """
        if name not in self._attr:
            raise Exception("Attribute does not exist")
        return self._attr[name]

    @abstractmethod
    def append(self,val):
        pass

class DataContainer(_BaseDataContainer):
    """
    A `DataContainer` is a container class for all elements of the same type in an instance (for example, vertices, edges or faces).
    It stores relevant information about the combinatorics (in the `_data` field) as well as various attributes onto the elements (in the `_attr` field)
    """

    def __init__(self, data : list = None, attributes : dict = None, id : str = ""):
        super().__init__(attributes, id)
        self._data = [] if data is None else list(data)
       
    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        try:
            self._data[key] = value
        except Exception as e:
            warnings.warn(f"Error in attribute '{key}' : {e}")
            raise Exception("Aborting")

    def __iter__(self):
        return self._data.__iter__()

    def __repr__(self):
        return self._data.__repr__()

    def __str__(self):
        return str(self._data) + "\nAttributes: "+str(self._attr.keys())

    def __len__(self):
        return len(self._data)

    @property
    def size(self) -> int:
        """The container size. Alias for `len(container)`

        Returns:
            int: number of elements in the container
        """
        return len(self._data)

    def empty(self) -> bool:
        """
        Returns:
            bool: whether the container is empty
        """
        return not self._data

    def clear(self):
        """Empty the container"""
        self._data = []
        self._attr = dict()

    def append(self, val):
        """Adds the value 'val' at the end of the container. Also expands all attributes to be able to receive a value for 'val'

        Args:
            val (Any): value to append to the container
        """
        self._data.append(val)
        for attr in self._attr.values():
            attr._expand(1)

    def __iadd__(self,other):
        """Performs several appends for a list of elements

        Args:
            other (Iterable|DataContainer): a collection of elements to append to the container

        Raises:
            Exception: fails if `other` is not an iterable

        Returns:
            self: self
        """
        if isinstance(other, list) or isinstance(other,tuple) or isinstance(other, set):
            self._data += list(other)
            for attr in self._attr.values():
                attr._expand(len(other))
        elif isinstance(other,DataContainer):
            self._data += other._data
            for attr in self._attr.values():
                attr._expand(other.n_elem)
        else:
            raise Exception("Could not append data container of type {} onto an attribute".format(type(other)))
        return self

class CornerDataContainer(_BaseDataContainer):
    """
    A `CornerDataContainer` is a variant of a DataContainer for corner elements. It is used for face corners, cell corners and cell faces. 
    
    Unlike a regular data container that stores a list of simplicial elements, a corner container stores two pieces of information: the associated vertex/face of the corner and the face/cell it belongs to.
    """
    def __init__(self, elem:list = None, adj:list = None, attributes : dict = None, id : str = ""):
        super().__init__(attributes, id)
        self._elem = [] if elem is None else list(elem)
        self._adj = [] if adj is None else list(adj)

    def __getitem__(self, key:int) -> int:
        """
        Shortcut for self.elem(key)

        Args:
            key (int): corner identifier

        Returns:
            int: the vertex or the face this corner points to
        """
        return self._elem[key]

    def element(self,key:int) -> int:
        """
        Returns the mesh element (vertex or face) associated with the corner 'key'

        Args:
            key (int): identifier of a corner

        Returns:
            int: identifier of a vertex or face
        """
        return self._elem[key]

    def adj(self, key:int) -> int:
        """
        Returns the mesh element (either face or cell) from which the corner 'key' belongs to

        Args:
            key (int): identifier of a corner

        Returns:
            int: identifier of a face or a cell
        """
        return self._adj[key]

    def __iter__(self):
        return self._elem.__iter__()

    def __repr__(self):
        return list(zip(self._elem, self._adj)).__repr__()

    def __str__(self):
        return str([x for x in zip(self._elem, self._adj)]) + "\nAttributes: "+str(self._attr.keys())

    @property
    def size(self):
        """The container size. Alias for `len(container)`

        Returns:
            int: number of elements in the container
        """
        return len(self._elem)

    def __len__(self):
        return len(self._elem)

    def empty(self):
        """
        Returns:
            bool: whether the container is empty
        """
        return not self._elem

    def clear(self):
        """Empty the container"""
        self._elem = []
        self._adj = []
        self._attr = dict()

    def append(self, val_elem, val_adj):
        self._elem.append(val_elem)
        self._adj.append(val_adj)
        for attr in self._attr.values():
            attr._expand(1)

    def __iadd__(self, other):
        if isinstance(other, list) or isinstance(other,tuple) or isinstance(other, set):
            for (e,a) in other:
                self._elem.append(e)
                self._adj.append(a)
            for attr in self._attr.values():
                attr._expand(len(other))
        elif isinstance(other, CornerDataContainer):
            self._elem += other._elem
            self._adj += other._adj
            for attr in self._attr.values():
                attr._expand(other.n_elem)
        else:
            raise Exception("Could not append data container of type {} onto an attribute".format(type(other)))
        return self


import numpy as np

class UnionFind(object):
    """Union-find disjoint sets datastructure.

    Original repository: [https://github.com/deehzee/unionfind/blob/master/unionfind.py](https://github.com/deehzee/unionfind/blob/master/unionfind.py)

    Union-find is a data structure that maintains disjoint set
    (called connected components or components in short) membership,
    and makes it easier to merge (union) two components, and to find
    if two elements are connected (i.e., belong to the same
    component).

    This implements the "weighted-quick-union-with-path-compression"
    union-find algorithm.  Only works if elements are immutable
    objects.

    Worst case for union and find: `(N + M log* N)`, with
    `N` elements and `M` unions. The function `log*` is the number of times needed to take `log` of a number until reaching 1. In practice, the amortized cost of
    each operation is nearly linear [1]_.

    Terms:
        **Component**:
            Elements belonging to the same disjoint set

        **Connected**:
            Two elements are connected if they belong to the same component.

        **Union**:
            The operation where two components are merged into one.

        **Root**:
            An internal representative of a disjoint set.

        **Find**:
            The operation to find the root of a disjoint set.

    Parameters:
        elements (None or container, optional):
            The initial list of elements.

    Attributes:
        n_elts (int): Number of elements.

        n_comps (int): Number of distjoint sets or components.
    
        __len__:
            Calling ``len(uf)`` (where ``uf`` is an instance of ``UnionFind``)
            returns the number of elements.

        __contains__:
            For ``uf`` an instance of ``UnionFind`` and ``x`` an immutable object,
            ``x in uf`` returns ``True`` if ``x`` is an element in ``uf``.

        __getitem__:
            For ``uf`` an instance of ``UnionFind`` and ``i`` an integer,
            ``res = uf[i]`` returns the element stored in the ``i``-th index.
            If ``i`` is not a valid index an ``IndexError`` is raised.

        __setitem__:
            For ``uf`` and instance of ``UnionFind``, ``i`` an integer and ``x``
            an immutable object, ``uf[i] = x`` changes the element stored at the
            ``i``-th index. If ``i`` is not a valid index an ``IndexError`` is
            raised.

    References:
        [1] http://algs4.cs.princeton.edu/lectures/

    """

    def __init__(self, elements=None):
        self.n_elts = 0  # current num of elements
        self.n_comps = 0  # the number of disjoint sets or components
        self._next = 0  # next available id
        self._elts = []  # the elements
        self._indx = {}  #  dict mapping elt -> index in _elts
        self._par = []  # parent: for the internal tree structure
        self._siz = []  # size of the component - correct only for roots

        if elements is None:
            elements = []
        for elt in elements:
            self.add(elt)


    def __repr__(self):
        return  (
            '<UnionFind:\n\telts={},\n\tsiz={},\n\tpar={},\nn_elts={},n_comps={}>'
            .format(
                self._elts,
                self._siz,
                self._par,
                self.n_elts,
                self.n_comps,
            ))

    def __len__(self):
        return self.n_elts

    def __contains__(self, x):
        return x in self._indx

    def __getitem__(self, index):
        if index < 0 or index >= self._next:
            raise IndexError('index {} is out of bound'.format(index))
        return self._elts[index]

    def __setitem__(self, index, x):
        if index < 0 or index >= self._next:
            raise IndexError('index {} is out of bound'.format(index))
        self._elts[index] = x

    def add(self, x):
        """Add a single disjoint element.

        Args:
            x (Any): immutable object
        """
        if x in self:
            return
        self._elts.append(x)
        self._indx[x] = self._next
        self._par.append(self._next)
        self._siz.append(1)
        self._next += 1
        self.n_elts += 1
        self.n_comps += 1

    def find(self, x) -> int:
        """Find the root of the disjoint set containing the given element.

        Args:
            x (Any): immutable object

        Returns:
            int: The (index of the) root.

        Raises:
            ValueError: If the given element is not found.
        """
        if x not in self._indx:
            raise ValueError('{} is not an element'.format(x))

        p = self._indx[x]
        while p != self._par[p]:
            # path compression
            q = self._par[p]
            self._par[p] = self._par[q]
            p = q
        return p

    def connected(self, x, y) -> bool:
        """
        Return whether the two given elements belong to the same component.
           
        Args:
            x (Any): immutable object
            y (Any): immutable object

        Returns:
            bool: True if x and y are connected, false otherwise.
        """
        return self.find(x) == self.find(y)

    def union(self, x, y):
        """
        Merge the components of the two given elements into one.

        Args:
            x (Any): immutable object
            y (Any): immutable object
        """
        # Initialize if they are not already in the collection
        for elt in [x, y]:
            if elt not in self:
                self.add(elt)

        xroot = self.find(x)
        yroot = self.find(y)
        if xroot == yroot:
            return
        if self._siz[xroot] < self._siz[yroot]:
            self._par[xroot] = yroot
            self._siz[yroot] += self._siz[xroot]
        else:
            self._par[yroot] = xroot
            self._siz[xroot] += self._siz[yroot]
        self.n_comps -= 1

    def component(self, x) -> set:
        """
        Find the connected component containing the given element.

        Args:
            x (Any): immutable object

        Raises:
            ValueError:  If the given element is not found.

        Returns:
            set: The component of x
        """
        if x not in self:
            raise ValueError('{} is not an element'.format(x))
        elts = np.array(self._elts)
        vfind = np.vectorize(self.find)
        roots = vfind(elts)
        return set(elts[roots == self.find(x)])

    def roots(self) -> set:
        """Return the set of roots of components
        
        Returns:
            set: A set of elements
        """
        return set(self.find(x) for x in self._elts)

    def components(self) -> list:
        """
        Return the list of connected components.

        Returns:
            list: A list of sets representing components
        """
        roots = self.roots()
        root_ids = dict((r,i) for i,r in enumerate(roots))
        components = [[] for _ in roots]
        for e in self._elts:
            i = root_ids[self.find(e)]
            components[i].append(e)
        return components

    def component_mapping(self) -> dict:
        """Return a dict mapping elements to their components.

        The returned dict has the following semantics

            `elt -> component containing elt`

        If x, y belong to the same component, the comp(x) and comp(y)
        are the same objects (i.e., share the same reference). Changing
        comp(x) will reflect in comp(y).  This is done to reduce
        memory.

        But this behaviour should not be relied on.  There may be
        inconsistency arising from such assumptions or lack thereof.

        If you want to do any operation on these sets, use caution.
        For example, instead of

        ```python
        s = uf.component_mapping()[item]
        s.add(stuff)
        # This will have side effect in other sets
        ```
        
        do
        ```python
        s = set(uf.component_mapping()[item]) # or
        s = uf.component_mapping()[item].copy()
        s.add(stuff)
        ```

        or
        ```python
        s = uf.component_mapping()[item]
        s = s | {stuff}  # Now s is different
        ```

        Returns:
            dict: A dict with the semantics: `elt -> component contianing elt`.

        """
        elts = np.array(self._elts)
        vfind = np.vectorize(self.find)
        roots = vfind(elts)
        distinct_roots = set(roots)
        comps = {}
        for root in distinct_roots:
            mask = (roots == root)
            comp = set(elts[mask])
            comps.update({x: comp for x in comp})
            # Change ^this^, if you want a different behaviour:
            # If you don't want to share the same set to different keys:
            # comps.update({x: set(comp) for x in comp})
        return comps




from ...mesh.datatypes import *
from .base import BaseParametrization
from ...geometry import Vec
from ... import geometry as geom
from ...attributes.glob import euler_characteristic

import numpy as np

from osqp import OSQP
import scipy.sparse as sp
from scipy.sparse import linalg

class LSCM(BaseParametrization):
    """
    Least-Square Conformal Map algorithm for computing a parametrization of a mesh.
    /!\\ The mesh should have the topology of a disk.
    Computed UVs are stored in the self.uvs container

    References:
        - [1] _Least Squares Conformal Maps for Automatic Texture Atlas Generation_, Levy et al. (2002)
        - [2] _Spectral Conformal Parameterization_, Mullen et al. (2008)
        - [3] _Intrinsic Parameterizations of Surface Meshes_, Desbrun et al. (2002)

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/lscm.py](https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/lscm.py)
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, verbose : bool=False, **kwargs):
        """
        Args:
            mesh (SurfaceMesh): the supporting mesh. Should be a surface with disk topology.
            verbose (bool, optional): verbose mode. Defaults to False.
            
        Keyword Args:
            eigen (bool, optional): whether to solve a linear system with two fixed points or use an eigen solver. Defaults to True
            save_on_corners (bool, optional): whether to store the results on face corners or vertices. Defaults to True
            solver_verbose (bool, optional): verbose level. Defaults to False.

        Attributes:
            uvs (Attribute): an attribute containing the uv-coordinates of the parametrization
            save_on_corners (bool): whether the coordinates are saved as a vertex attribute or a face_corner attribute
            residual (float): the least-square residual
        """
        super().__init__("LSCM", mesh, verbose=verbose, **kwargs)
        self.residual : float = None # Final value of LSCM energy (residual of least square)
        self._eigen = kwargs.get("eigen", True)
        self._solver_verbose = kwargs.get("solver_verbose", False)

    def run(self):
        """
        Calls the solver on the LSCM system.

        Raises:
            Exception: fails if the mesh is not a topological disk
        """
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run LSCM.")
        if self._eigen:
            U = self._solve_eigen()
        else:
            U = self._solve(verbose=self._solver_verbose)
        U = U.reshape((U.size//2, 2))
        U = self._scale(U)        
        # Retrieve uvs and write them in attribute
        if self.save_on_corners:
            self.uvs = self.mesh.face_corners.create_attribute("uv_coords", float, 2, dense=True)
            for c,v in enumerate(self.mesh.face_corners):
                self.uvs[c] = Vec(U[v])
        else:
            self.uvs = self.mesh.vertices.create_attribute("uv_coords", float, 2, dense=True)
            for v in self.mesh.id_vertices:
                self.uvs[v] = Vec(U[v])

    def _build_system(self):
        n = len(self.mesh.vertices)
        m = len(self.mesh.faces)
        cols, rows, vals = (np.zeros(12*m) for _ in range(3))
        for t,(i,j,k) in enumerate(self.mesh.faces):
            Pi, Pj, Pk = (self.mesh.vertices[_v] for _v in (i,j,k))
            X,Y,_ = geom.face_basis(Pi,Pj,Pk)
            Qi, Qj, Qk = ( Vec(X.dot(_P), Y.dot(_P)) for _P in (Pi,Pj,Pk))
            
            for _i, (r,c,v) in enumerate([
                # first row
                (2*t, 2*i, Qk[0] - Qj[0]),
                (2*t, 2*j, Qi[0] - Qk[0]),
                (2*t, 2*k, Qj[0] - Qi[0]),
                (2*t, 2*i+1, Qk[1] - Qj[1]),
                (2*t, 2*j+1, Qi[1] - Qk[1]),
                (2*t, 2*k+1, Qj[1] - Qi[1]),
                # second row
                (2*t+1, 2*i, Qk[1] - Qj[1]),
                (2*t+1, 2*j, Qi[1] - Qk[1]),
                (2*t+1, 2*k, Qj[1] - Qi[1]),
                (2*t+1, 2*i+1, Qj[0] - Qk[0]),
                (2*t+1, 2*j+1, Qk[0] - Qi[0]),
                (2*t+1, 2*k+1, Qi[0] - Qj[0])
            ]):
                rows[12*t+_i], cols[12*t+_i], vals[12*t+_i] = r, c, v 
        A = sp.csc_matrix((vals, (rows, cols)),  shape=(2*m,2*n))
        self.log("Size of system :", A.shape)
        self.log(f"({A.nnz} non zero values)")
        return A

    def _solve(self, verbose:bool):
        self.log("Building system...")
        n = len(self.mesh.vertices)
        A = self._build_system()

        # index locked 1 has value (0,0) and index locked2 has value (1,1)
        e = self.mesh.boundary_edges[0]
        locked1, locked2 = self.mesh.edges[e]
        locked1, locked2 = min(locked1, locked2), max(locked1, locked2)

        instance = OSQP()
        Q = A.transpose() @ A
        Cst = sp.lil_matrix((4,2*n))
        Cst[0, 2*locked1] = 1
        Cst[1, 2*locked1+1] = 1
        Cst[2, 2*locked2] = 1
        Cst[3, 2*locked2+1] = 1
        Cst = Cst.tocsc()
        Cst_rhs = np.array([0.,0.,1.,0.])
        
        self.log("Solving system...")
        instance.setup(Q, A=Cst, l=Cst_rhs, u=Cst_rhs, eps_abs=0., verbose=verbose)
        res = instance.solve()
        self.log(f"System solved in {res.info.run_time} s")
        self.residual = res.info.obj_val
        self.log(f"Residual: {self.residual}")
        return res.x

    def _solve_eigen(self):
        self.log("Building system...")
        A = self._build_system()
        self.log("Solving for eigenvector...")
        # u,s,v = linalg.svds(A, 2, which="SM", solver="lobpcg", maxiter=1000)
        # U = v[0]
        eigs, U = linalg.eigsh(A.transpose() @ A, 2, which="SM", tol=1e-2)
        self.log("Smallest eigenvalues:", eigs)
        U = U[:,1]
        return U

    def _scale(self, U):
        # scale UVs in bounding box [0;1]^2
        xmin,xmax,ymin,ymax = float("inf"), -float("inf"), float("inf"), -float("inf")
        for i in range(U.shape[0]):
            u,v = U[i][0], U[i][1]
            xmin = min(xmin, u)
            xmax = max(xmax, u)
            ymin = min(ymin, v)
            ymax = max(ymax, v)
        scale_x = xmax-xmin
        scale_y = ymax-ymin
        scale = min(scale_x, scale_y)
        # apply scale
        return U/scale


from ..utils import Logger
from abc import ABC, abstractmethod

class Worker(Logger, ABC):

    def __init__(self, name:str, verbose:bool = False):
        super().__init__(name, verbose)

    @abstractmethod
    def run(self, *args, **kwargs):
        """A M.Worker implements a `.run()` method where the algorithm is run on an input mesh"""
        pass

    def __call__(self, *args, **kwargs):
        self.run(*args, **kwargs)
        return self


from scipy.optimize import fmin_l_bfgs_b
import scipy.sparse as sp
import numpy as np
from numba import njit, prange

from .base import BaseParametrization
from ...mesh import copy
from ...mesh.datatypes import *
from ...mesh.mesh_attributes import ArrayAttribute
from ... import geometry as geom
from ...utils.argument_check import check_argument
from ...attributes.glob import euler_characteristic
from ...attributes.misc_corners import cotangent
from ...operators import area_weight_matrix

@njit(cache=True, fastmath=True)
def regul(x, eps=1e-18):
    return (x*x + eps)**(0.25)

@njit(cache=True, fastmath=True)
def regul_prime(x, eps=1e-18):
    y = x*x + eps
    return 0.5 * x / (y**0.75)

@njit(cache=True, parallel=True)
def cotan_embedding_energy_and_gradient(X, tris, locked):
    E = 0
    grad = np.zeros_like(X)
    nt = tris.shape[0]
    for t in prange(nt):
        A,B,C = tris[t,:]
        iuA,ivA,iuB,ivB,iuC,ivC = 2*A, 2*A+1, 2*B, 2*B+1, 2*C, 2*C+1
        uA,vA,uB,vB,uC,vC = X[iuA], X[ivA], X[iuB], X[ivB], X[iuC], X[ivC]

        At = uA * (vB - vC) + uB * (vC - vA) + uC * (vA - vB)
        grad[iuA] += 2*(vB - vC)
        grad[ivA] += 2*(uC - uB)
        grad[iuB] += 2*(vC - vA)
        grad[ivB] += 2*(uA - uC)
        grad[iuC] += 2*(vA - vB)
        grad[ivC] += 2*(uB - uA)
        
        l_bc = (uB-uC)**2 + (vB-vC)**2
        l_ac = (uA-uC)**2 + (vA-vC)**2  
        l_ab = (uA-uB)**2 + (vA-vB)**2

        L_ab = - l_ab + l_bc + l_ac 
        L_bc =   l_ab - l_bc + l_ac
        L_ac =   l_ab + l_bc - l_ac
        U = 2*l_bc*l_ac + 2*l_ac*l_ab + 2*l_bc*l_ab - l_bc*l_bc - l_ac*l_ac - l_ab*l_ab

        Dt = regul(U)
        Dtp = regul_prime(U)
        grad[iuA] += 4*Dtp * ( (uA - uB)*L_ab + (uA - uC)*L_ac )
        grad[ivA] += 4*Dtp * ( (vA - vB)*L_ab + (vA - vC)*L_ac )
        grad[iuB] += 4*Dtp * ( (uB - uA)*L_ab + (uB - uC)*L_bc )
        grad[ivB] += 4*Dtp * ( (vB - vA)*L_ab + (vB - vC)*L_bc )
        grad[iuC] += 4*Dtp * ( (uC - uA)*L_ac + (uC - uB)*L_bc )
        grad[ivC] += 4*Dtp * ( (vC - vA)*L_ac + (vC - vB)*L_bc )
        E += (Dt + 2*At)
    grad[locked] = 0.
    return E, grad

class CotanEmbedding(BaseParametrization):
    """
    Given a parametrization of a disk where boundary is fixed, we can optimize the difference between unsigned and signed areas of triangles to compute a parametrization that is foldover-free.

    Warning: 
        The input mesh should have the topology of a disk.

        UV coordinates are computed per vertex and not per corner. See `mouette.attributes.scatter_vertices_to_corners` for conversion.

    References:
        [1] _Embedding a triangular graph within a given boundary_, Xu et al. (2011)

    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/cotan_embedding.py](https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/cotan_embedding.py)
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, uv_init: ArrayAttribute, mode: str = "bfgs", verbose:bool=False, **kwargs):
        """
        Args:
            mesh (SurfaceMesh): the supporting mesh.
            uv_init (ArrayAttribute): array of initial uv-coordinates per vertices. np.array of shape (V,2) or mouette.ArrayAttribute object.
            verbose (bool, optional): verbose mode. Defaults to False.
            mode (str): optimization mode. Either "bfgs" or "alternate". Defaults to "bfgs".
        
        Keyword Args:
            tutte_if_convex (bool): if True and if the boundary of the shape is convex and mode=="alternate", will simply runs Tutte's embedding. Defaults to True
            solver_verbose (bool): solver verbose mode. Defaults to False.
        """
        kwargs["uv_attr"] = uv_init
        super().__init__("CotanEmbedding", mesh, verbose=verbose, **kwargs)
        self._save_on_corners = False
        self._mode = mode
        check_argument("mode", self._mode, str, ["bfgs", "alternate"])
        self._tutte_if_convex : bool = kwargs.get("tutte_if_convex", True)
        self._solver_verbose : bool = kwargs.get("solver_verbose", False)

    def run(self):
        """
        Runs the optimization

        Raises:
            Exception: Fails if the mesh is not a topological disk
            Exception: Fails if the mesh is not triangular
        """
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run parametrization.")
        if not self.mesh.is_triangular():
            raise Exception("Mesh is not triangular.")
        
        # Call corresponding optimizer
        if self._mode=="bfgs":
            # Initialize variable vector from uvs
            var = np.concatenate([self.uvs[v] for v in self.mesh.id_vertices])
            locked = np.array([self.mesh.is_vertex_on_border(v) for v in self.mesh.id_vertices])
            locked = np.repeat(locked, 2)
            triangles = np.array(self.mesh.faces)
            self.log("Optimize cotan energy")
            var,energy,infos = fmin_l_bfgs_b(
                cotan_embedding_energy_and_gradient, # function and gradient to optimize
                var, # Initial variables
                args = (triangles, locked), # arguments to be given to the function to optimize
                maxiter = 1e4, # maximum number of iterations
                m = 11, # number of inner iterations for hessian approx.
                maxls = 30, # maximum number of linesearch
                factr = 10, # stops iterations if improvement is < factr*eps_machine
                disp=self._solver_verbose, # verbose level
            )
            self.log(f"Stopped after {infos['nit']} iterations")
            self.log(f"Final energy : {energy}")
            for v in self.mesh.id_vertices:
                self.uvs[v] = var[2*v:2*v+2]    

        elif self._mode=="alternate":
            var = self.uvs.as_array(len(self.mesh.vertices))
            var = self._optimize_alternate(var, self._tutte_if_convex, self._solver_verbose)
            for v in self.mesh.id_vertices:
                self.uvs[v] = var[v,:]

    def _optimize_alternate(self, var, tutte, verbose):

        def check_dets(var):
            for (A,B,C) in self.mesh.faces:
                uA,vA = var[A,:]
                uB,vB = var[B,:]
                uC,vC = var[C,:]
                if (uA * (vC - vB) + uB * (vA - vC) + uC * (vB - vA)) < 1e-8 : return False
            return True

        n_iter = 0
        n_iter_max = 100
        n_coeff_lap = 12*len(self.mesh.faces)
        rows = np.zeros(n_coeff_lap, dtype=np.int32)
        cols = np.zeros(n_coeff_lap, dtype=np.int32)
        coeffs = np.zeros(n_coeff_lap, dtype=np.float64)
        Iinds = np.array(self.mesh.interior_vertices, dtype=np.int32)
        Binds = np.array(self.mesh.boundary_vertices, dtype=np.int32)
        AreaW = area_weight_matrix(self.mesh, inverse=True).tocsc()
        if tutte:
            cot = cotangent(self.mesh, persistent=False)

        while n_iter<n_iter_max and not check_dets(var):
            n_iter += 1
            self.log(f"Iter {n_iter}/{n_iter_max}")
            # build cotan laplacian  
            _c = 0
            eps = 1e-10 if n_iter<10 else 0
            for iF,(A,B,C) in enumerate(self.mesh.faces):
                if tutte and n_iter==1:
                    cot_a,cot_b,cot_c = (cot[3*iF+_i]/2 for _i in range(3))
                else:
                    uA,vA = var[A,:]
                    uB,vB = var[B,:]
                    uC,vC = var[C,:] 
                    l_bc = (uB-uC)**2 + (vB-vC)**2
                    l_ac = (uA-uC)**2 + (vA-vC)**2  
                    l_ab = (uA-uB)**2 + (vA-vB)**2
                    Eli = 2*l_bc*l_ac + 2*l_ac*l_ab + 2*l_bc*l_ab - l_bc*l_bc - l_ac*l_ac - l_ab*l_ab
                    if eps==0:
                        El = np.sqrt(Eli)
                    else:
                        El = regul(Eli, eps)
                    cot_a = (l_ac+l_ab-l_bc)/El
                    cot_b = (l_bc+l_ab-l_ac)/El
                    cot_c = (l_bc+l_ac-l_ab)/El
                for (i, j, v) in [(A, B, cot_c/2), (B, C, cot_a/2), (C, A, cot_b/2)]:
                    rows[_c], cols[_c], coeffs[_c], _c = i, i, v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = j, j, v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = i, j, -v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = j, i, -v, _c+1 
            L = AreaW * sp.csc_matrix((coeffs,(rows,cols)), dtype= np.float64)
            LI = L[Iinds,:][:,Iinds]
            LB = L[Iinds,:][:,Binds] 
            # solve linear system
            var[Iinds,0] = sp.linalg.spsolve(LI, -LB.dot(var[Binds,0])) # first system for u
            var[Iinds,1] = sp.linalg.spsolve(LI, -LB.dot(var[Binds,1])) # second system for v
        return var
    
    @property
    def flat_mesh(self) -> SurfaceMesh:
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None: return None
        flat = copy(self.mesh)
        # build the flat mesh : vertex coordinates are uv of original mesh
        for T in self.mesh.id_faces:
            for i,v in enumerate(self.mesh.faces[T]):
                flat.vertices[v] = geom.Vec(self.uvs[v][0], self.uvs[v][1], 0.)
        return flat


from ..attributes import face_barycenter, face_circumcenter
from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from ..mesh.mesh import _instanciate_raw_mesh_data

@allowed_mesh_types(SurfaceMesh)
def dual_mesh(mesh: SurfaceMesh, mode:str = "barycenter") -> SurfaceMesh:
    """Computes the dual mesh of a mesh

    Args:
        mesh (SurfaceMesh): input surface
        mode (str): which position

    Returns:
        SurfaceMesh
    """
    out = RawMeshData()
    if mode.lower() == "barycenter":
        dual_pts = face_barycenter(mesh, persistent=False)
    elif mode.lower() == "circumcenter":
        dual_pts = face_circumcenter(mesh)
    for F in mesh.id_faces:
        out.vertices.append(dual_pts[F])
    for V in mesh.id_vertices:
        out.faces.append(mesh.connectivity.vertex_to_faces(V))
    return _instanciate_raw_mesh_data(out, 2)


from abc import ABC, abstractmethod
from collections import deque
from ...mesh.datatypes import *
from ...mesh import merge

class SpanningTree(ABC):
    """A spanning tree defined over the connectivity of a mesh"""

    def __init__(self, mesh : Mesh):
        self.mesh : Mesh = mesh

        self.root = None

        self.parent : list = None # v -> parent of v
        self.children : list = None # v -> list of indices of children of v (inverse of self.parent)
        self.edges : list = None # list of (u,v) with u<v and u parent of v (or v parent of u)

        self._computed : bool = False

    def __call__(self):
        self.compute()
        return self

    @abstractmethod
    def compute(self) -> None :
        """Runs the traversal and builds the tree"""
        self._computed = True

    def traverse(self, order="BFS"):
        """
        Iterator on the nodes of a tree. 
        Returns tuple of form (node, parent)
        
        Parameters:
            order (str, optional):  BFS or DFS order. Defaults to "BFS".
        """
        if not self._computed:
            raise Exception("Tree was not computed. Call the .compute() method first")

        if order not in ["BFS", "DFS"]:
            raise Exception("Error when traversing a spanning tree: given 'order' argument should be 'DFS' or 'BFS'. Got {}".format(order))
            
        isBFS = (order=="BFS")
        queue = deque()
        def pop():
            if isBFS: return queue.popleft()
            return queue.pop()
        queue.append((self.root, None))
        while len(queue)>0:
            node,parent = pop()
            yield node,parent
            for child in self.children[node]:
                queue.append((child,node))

    @abstractmethod
    def build_tree_as_polyline(self) -> PolyLine : 
        pass

class SpanningForest(ABC):
    """
    Unlike SpanningTree, SpanningForest allows to define several roots and is guaranteed to visit every element of a mesh.
    """

    @forbidden_mesh_types(PointCloud)
    def __init__(self, mesh : Mesh):
        self.mesh  : Mesh = mesh
        self.trees : list = [] # a list of SpanningTree objects
        self.roots : list = []

    def __call__(self):
        self.compute()
        return self
        
    @property
    def n_trees(self) -> int:
        """Number of trees in the forest
        
        Returns:
            int:
        """
        return len(self.trees)

    def __getitem__(self, key):
        return self.trees[key]

    @property
    def edges(self) -> list:
        """List of edges in the trees

        Returns:
            list: list of edges
        """
        _edges = []
        for tree in self.trees:
            _edges += tree.edges
        return _edges

    @abstractmethod
    def compute(self) -> None : 
        pass

    def traverse(self, order="BFS"):
        """
        Iterator on the nodes of a tree 
        
        Parameters:
            order (str, optional):  BFS or DFS order. Defaults to "BFS".
        """
        for tree in self.trees:
            for el in tree.traverse(order=order):
                yield el

    def build_tree_as_polyline(self) -> PolyLine :
        """Builds the tree as a new polyline object. Useful for debug and visualization purposes

        Returns:
            PolyLine: the tree
        """
        trees = [t.build_tree_as_polyline() for t in self.trees]
        return merge(trees)


from .mesh.datatypes import *
from .mesh import from_arrays
from .attributes import face_normals, face_area, edge_length
import numpy as np
from numpy.random import random, choice
from .geometry import Vec, AABB
from .utils import check_argument

def sample_sphere(center : Vec, radius : float, n_pts : int, return_point_cloud : bool = False):
    """Samples points uniformly on the surface of a 3D sphere

    Args:
        center (Vec): center of the sphere
        radius (float): radius of the spheres
        n_pts (int): number of points to sample
        return_point_cloud (bool, optional): whether to compile the points in a PointCloud object or return the raw numpy array. Defaults to False.

    Returns:
        PointCloud | np.ndarray: a sampled point cloud of `n_pts` points
    """
    pts = np.vstack([
        np.random.normal(0.,1., size=n_pts),
        np.random.normal(0.,1., size=n_pts),
        np.random.normal(0.,1., size=n_pts)
    ]).T
    pts /= np.linalg.norm(pts, axis=1, keepdims=True) # uniform distribution of the unit sphere
    pts = radius*pts + center
    if return_point_cloud:
        pointcloud = PointCloud()
        pointcloud.vertices += list(pts)
        return pointcloud
    else:
        return pts

def sample_ball(center : Vec, radius : float, n_pts : int, return_point_cloud : bool = False):
    """Samples points uniformly inside a 3D ball.

    Args:
        center (Vec): center of the ball
        radius (float): radius of the ball
        n_pts (int): number of points to sample
        return_point_cloud (bool, optional): whether to compile the points in a PointCloud object or return the raw numpy array. Defaults to False.

    Returns:
        PointCloud | np.ndarray: a sampled point cloud of `n_pts` points
    """
    pts = np.vstack([
        np.random.normal(0.,1., size=n_pts),
        np.random.normal(0.,1., size=n_pts),
        np.random.normal(0.,1., size=n_pts)
    ]).T
    pts /= np.linalg.norm(pts, axis=1, keepdims=True)
    R = np.random.uniform(0, radius, n_pts).reshape((n_pts,1))
    R = np.cbrt(R) # R^{1/3} for uniform distribution
    pts = pts*R + center
    if return_point_cloud:
        pointcloud = PointCloud()
        pointcloud.vertices += list(pts)
        return pointcloud
    else:
        return pts

def sample_AABB(
        box : AABB, 
        n_pts : int,
        mode : str = "uniform",
        return_point_cloud : bool = False
    ):
    """Sample a point cloud uniformly at random inside an axis-aligned bounding box. Works by sampling the unit cube and applying an affine transformation.

    Args:
        box (AABB): the domain of sampling
        n_pts (int): number of points to sample. If mode is 'grid', the function may return a slightly smaller number of points (nearest perfect n-th root).
        mode (str): sampling mode. 'uniform' or 'grid'. Uniform takes points at random, while 'grid' generates a grid of regularly spaced points.
        return_point_cloud (bool, optional): whether to compile the points in a PointCloud object or return the raw numpy array. Is ignored if the bounding box has dimension >3. Defaults to False.

    Raises:
        Exception: fails if the bounding box is empty.
        ValueError: fails if the dimension of the box is > 3 and return_point_cloud is set to True, so that no PointCloud object with dim > 3 is created.

    Returns:
        PointCloud | np.ndarray: a sampled point cloud of `n_pts` points
    """
    check_argument("mode", mode, str, ["uniform", "grid"])
    if box.is_empty():
        raise Exception(f"Received AABB {box}, is empty. Cannot sample valid points.")
    if box.dim>3 and return_point_cloud:
        raise ValueError(f"No PointCloud can be generated for a bounding box of dimension {box.dim}>3.")
    if mode=="uniform":
        points = box.mini + box.span * random((n_pts,box.dim))
    elif mode=="grid":
        res = round(np.power(n_pts, 1/box.dim))
        Xdims = (np.linspace(0,1,res) for _ in range(box.dim))
        points = np.vstack(list(map(np.ravel, np.meshgrid(*Xdims)))).T
    if return_point_cloud:
        return from_arrays(points)
    else:
        return points

@allowed_mesh_types(PolyLine)
def sample_polyline(
    mesh : PolyLine, 
    n_pts : int,
    return_point_cloud : bool = False
    ):
    """
    Sample a point cloud uniformly at random from a polyline

    Args:
        mesh (PolyLine): the input polyline
        n_pts (int): number of points to sample
        return_point_cloud (bool, optional): whether to compile the points in a PointCloud object or return the raw numpy array. Defaults to False.

    Returns:
        PointCloud | np.ndarray:  a sampled point cloud of `n_pts` points
    """
    NE = len(mesh.edges)
    sampled_pts = np.zeros((n_pts, 3))
    if NE>1:
        lengths = edge_length(mesh, persistent=False).as_array()
        lengths /= np.sum(lengths)
        edges = choice(NE, size=n_pts, p=lengths)
    else:
        edges = [0]*n_pts
    for i,e in enumerate(edges):
        pA,pB = (mesh.vertices[_v] for _v in mesh.edges[e])
        t = np.random.random()
        sampled_pts[i,:] = t*pA + (1-t)*pB
    if return_point_cloud:
        pointcloud = PointCloud()
        pointcloud.vertices += list(sampled_pts)
        return pointcloud
    else:
        return sampled_pts

@allowed_mesh_types(SurfaceMesh)
def sample_surface(
    mesh : SurfaceMesh, 
    n_pts : int, 
    return_point_cloud : bool = False, 
    return_normals : bool=False
    ):
    """
    Sample a point cloud uniformly at random from a surface mesh

    Args:
        mesh (SurfaceMesh): input mesh
        n_pts (int): number of points to sample
        return_point_cloud (bool, optional): whether to compile the points in a PointCloud object or return the raw numpy array. Defaults to False.
        return_normals (bool, optional): wether to assign the normal of the faces to sampled points. Only has effect if return_point_cloud is set to True. Defaults to False.

    Returns:
        PointCloud | np.ndarray: a sampled point cloud of `n_pts` points
        np.ndarray : the associated normals (if sample_normals was set to True)
    """
    assert mesh.is_triangular()
    NF = len(mesh.faces)
    areas = face_area(mesh, persistent=False).as_array()
    areas /= np.sum(areas)
    sampled_pts = np.zeros((n_pts,3))
    
    sampled_faces = choice(NF, size=n_pts, p=areas) # faces on which we take the points (probability weighted by area)
    if return_normals:
        normals = face_normals(mesh,persistent=False)
        sampled_normals = np.array([normals[f] for f in sampled_faces])

    for i,f in enumerate(sampled_faces):
        pA,pB,pC = (mesh.vertices[_v] for _v in mesh.faces[f])
        u1,u2 = random(2)
        r1 = 1 - np.sqrt(u1)
        r2 = u2*(1-r1)
        sampled_pts[i,:] = pA+ r1*(pB-pA) + r2*(pC-pA)
    
    if return_point_cloud:
        pointcloud = PointCloud()
        pointcloud.vertices += list(sampled_pts)
        if return_normals: 
            pc_normals = pointcloud.vertices.create_attribute("normals", float, 3, dense=True)
            pc_normals._data = sampled_normals
        return pointcloud
    else:
        if return_normals:
            return sampled_pts,sampled_normals
        return sampled_pts


import numpy as np

class Vec(np.ndarray):
    """A simple class to manipulate vectors in mouette. 
    Basically, it inherits from a numpy array and implements some quality of life features for 2D and 3D vectors especially.
    """
    ###### Constructors ######
    def __new__(cls, *a):
        if len(a)==1:
            obj = np.asarray(a[0]).view(cls)
        else:
            obj = np.asarray(a).view(cls)
        return obj
    
    @classmethod
    def from_complex(cls, c : complex) -> "Vec":
        """2D vector from complex number

        Args:
            c (complex):

        Returns:
            Vec: `Vec(c.real, c.imag)`
        """
        return cls(c.real, c.imag)

    @classmethod
    def random(cls, n : int) -> "Vec":
        """Generates a random vector of size `n` with coefficients sampled uniformly and independently in [0;1)

        Args:
            n (int): size

        Returns:
            Vec:
        """
        return cls(np.random.random(n))

    @classmethod
    def zeros(cls, n : int) -> "Vec":
        """Generates a vector of size `n` full of zeros.

        Args:
            n (int): size

        Returns:
            Vec:
        """
        return cls(np.zeros(n))
    
    @classmethod
    def X(cls) -> "Vec":
        """The [1,0,0] vector

        Returns:
            Vec: [1,0,0]
        """
        return Vec(1.,0.,0.)

    @classmethod
    def Y(cls) -> "Vec":
        """The [0,1,0] vector

        Returns:
            Vec: [0,1,0]
        """
        return Vec(0.,1.,0.)

    @classmethod
    def Z(cls) -> "Vec":
        """The [0,0,1] vector

        Returns:
            Vec: [0,0,1]
        """
        return Vec(0.,0.,1.)

    ###### Accessors ######
    @property
    def x(self) -> float: 
        """First coordinate of the vector

        Returns:
            float: `vec[0]`
        """
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value

    @property
    def y(self) -> float:
        """Second coordinate of the vector

        Returns:
            float: `vec[1]`
        """
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value

    @property
    def xy(self) -> np.ndarray:
        """First two coordinates of the vector

        Returns:
            np.ndarray[float]: `vec[:2]`
        """
        return self[:2]

    @property
    def z(self) -> float:
        """Third coordinate of the vector

        Returns:
            float: `vec[3]`
        """
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value

    ###### Methods ######
    def norm(self, which="l2") -> float :
        """Vector norm. Three norms are implemented: the Euclidean l2 norm, the l1 norm or the l-infinite norm:

        l2 : $ \\sqrt{ \\sum_i v[i]^2 } $

        l1 : $ \\sum_i |v[i]| $

        linf : $ \\max_i |v[i]| $

        Args:
            which (str, optional): which norm to compute. Choices are "l2", "l1" and "linf". Defaults to "l2".

        Returns:
            float: the vector's norm
        """
        if which=="l2":
            return np.sqrt(np.dot(self, self))
        elif which=="l1":
            return np.sum(np.abs(self))
        elif which=="linf":
            return np.max(np.abs(self))

    def dot(self, other) -> float:
        """Dot product between two vectors:

        $ a \\cdot b = \\sum_i a[i]b[i]$

        Args:
            other (Vec): other vector to dot with

        Returns:
            float: the dot product
        """
        return np.dot(self, other)

    def outer(self,other) -> np.ndarray:
        """Outer product between two vectors:

        $a \\otimes b = c \\in \\mathbb{R}^{n \\times n}$ such that $c[i,j] = a[i]b[j]$.

        Args:
            other (Vec): the second vector

        Returns:
            np.array: an array of shape (n,n)
        """
        return np.outer(self,other)

    def normalize(self, which="l2"):
        """Normalizes the vector to have unit norm.

        Args:
            which (str, optional): which norm to compute. Choices are "l2", "l1" and "linf". Defaults to "l2".
        """
        self /= self.norm(which)

    @staticmethod
    def normalized(vec, which="l2") -> "Vec":
        """Computes and returns a normalized vector of the input vector `vec`.

        Args:
            vec (Vec): the input vector
            which (str, optional): which norm to compute. Choices are "l2", "l1" and "linf". Defaults to "l2".

        Returns:
            Vec: the normalized vector
        """
        nrm = Vec.norm(vec, which)
        np.seterr(all='raise')
        # we want the following to fail when a division by zero is encountered
        out = Vec(vec/nrm)
        np.seterr(all='warn')
        return out


from .vector import Vec
from scipy.spatial.transform import Rotation
import math
from .. import geometry as geom

def rotate_2d(v : Vec, angle : float) -> Vec:
    """Rotates a 2D vector in the plane

    Parameters:
        v (Vec): the vector to rotate
        angle (float): angle of rotation in radiants

    Returns:
        Vec: the rotated vector
    """
    ca,sa = math.cos(angle), math.sin(angle)
    v2 = Vec(0.,0.)
    v2.x = v[0] * ca - v[1] * sa
    v2.y = v[0] * sa + v[1] * ca
    return v2

def rotate_around_axis(inp : Vec, _axis : Vec, angle : float) ->  Vec:
    c,s = math.cos(angle), math.sin(angle)
    inp = Vec(inp)
    axis = Vec.normalized(_axis)
    if abs(angle)<1E-12 or axis.norm()<1E-12: return inp
    out = Vec(0.,0.,0.)
    u,v,w = axis
    out.x = (c + u*u*(1-c))   * inp.x + (u*v*(1-c) - w*s) * inp.y + (u*w*(1-c) + v*s) * inp.z
    out.y = (u*v*(1-c) + w*s) * inp.x + (c + v*v*(1-c))   * inp.y + (v*w*(1-c) - u*s) * inp.z
    out.z = (u*w*(1-c) - v*s) * inp.x + (v*w*(1-c) + u*s) * inp.y + (c + w*w*(1-c))   * inp.z
    return out

def axis_rot_from_z(v : Vec) -> Vec:
    """The (smallest) rotation to align the z axis (0,0,1) with the vector v
    """
    Z = Vec(0.,0.,1.)
    axis = geom.cross(Z, v)
    angle = geom.angle_2vec3D(Z, v)
    if axis.norm()>1e-8: axis = Vec.normalized(axis) * angle
    return axis

def match_rotation(Ra : Rotation, Rb : Rotation, symgroup = Rotation.create_group("O"), threshold=math.pi/4):
    """Given two rotation matrices, compute the matching between the two: 
    the minimal rotation going from Ra to s(Rb) with s in some symmetry group (most often the octahedral group for cube symmetries)

    Parameters:
        Ra (Rotation)
        Rb (Rotation)
        symgroup (Rotation) : a rotation group (given by the `scipy.spatial.transform.Rotation.create_group` method). 
            Defaults to the octahedral group (24 direct symmetries of the cube)

    Returns:
        Rotation: the minimal rotation going from Ra to Rb
    """
    RaT = Ra.inv()
    Rab = Rb * RaT
    if Rab.magnitude()<threshold: return Rab # no need to perform a symmetry
    #return Rab.reduce(left=symgroup)
    min_w = float("inf")
    min_r = None
    for S in symgroup:
        match = Rb * S * RaT
        w = match.magnitude()
        if w<threshold: return match
        if w<min_w:
            min_w = w
            min_r = match
    return min_r


from .base import Mesh
from .pointcloud import PointCloud
from .linear import PolyLine
from .surface import SurfaceMesh
from .volume import VolumeMesh
from .type_checks import allowed_mesh_types, forbidden_mesh_types


from ..datatypes import *
from ..mesh_data import RawMeshData
from collections import deque
import warnings

def parse_field(data : deque, container, nlines, nelem):
    for _ in range(nlines):
        line = data.popleft().split()
        d = [int(u.strip())-1 for u in line][:nelem]
        container.append(d)

def import_medit(path):
    obj = RawMeshData()
    data = deque()
    with open(path, 'r' ) as meditf:
        data = deque([x.strip() for x in meditf.readlines()])
                
    while data:
        line = data.popleft()

        if line=="End": break # end of file

        elif line=="Vertices":
            nv = int(data.popleft())
            for _ in range(nv):
                line = data.popleft().split()
                vertex = [float(u.strip()) for u in line[:3]]
                obj.vertices.append(vertex)

        elif line=="Edges":
            ne = int(data.popleft())
            parse_field(data, obj.edges, ne, 2)

        elif line=="Triangles":
            nt = int(data.popleft())
            parse_field(data, obj.faces, nt, 3)

        elif line=="Quadrilaterals":
            nq = int(data.popleft())
            parse_field(data, obj.faces,  nq, 4)

        elif line=="Tetrahedra":
            nc = int(data.popleft())
            parse_field(data, obj.cells, nc, 4)
        
        elif line=="Hexahedra":
            nc = int(data.popleft())
            parse_field(data, obj.cells, nc, 6)

    return obj

def count_cells(mesh : RawMeshData):
    cell_count = [0,0,0]
    # 0 = hexahedra
    # 1 = tetrahedra
    # 2 = else
    for c in mesh.cells:
        N = len(c)
        if N==8: cell_count[0] += 1
        elif N==4 : cell_count[1] += 1
        else : cell_count[2] += 1
    return cell_count

def count_faces(mesh : RawMeshData):
    face_count = [0, 0, 0]
    # 0 = quad
    # 1 = tri
    # 2 = else
    for f in mesh.faces:
        if len(f)==4 : face_count[0] +=1
        elif len(f)==3 : face_count[1] +=1
        else: face_count[2] += 1
    return face_count

def export_medit(mesh : RawMeshData, path):
    with open(path, 'w') as f:
        f.write("MeshVersionFormatted 1\n")
        f.write("Dimension 3\n") # does not matter -> 3 as default

        if not mesh.vertices.empty():
            f.write("Vertices\n")
            f.write("{}\n".format(len(mesh.vertices)))
            for v in mesh.vertices:
                f.write("{} {} {} 1\n".format(v[0], v[1], v[2]))
            f.write("\n")

        if hasattr(mesh, "edges") and not mesh.edges.empty():
            f.write("Edges\n")
            if mesh.edges.has_attribute("hard_edges"):
                f.write("{}\n".format(len(mesh.edges.get_attribute("hard_edges"))))
                for e in mesh.edges.get_attribute("hard_edges"):
                    a,b = mesh.edges[e]
                    f.write("{} {} 1\n".format(a+1,b+1))
            else:
                f.write("{}\n".format(len(mesh.edges)))
                for a,b in mesh.edges:
                    f.write("{} {} 1\n".format(a+1,b+1))
            f.write("\n")
        
        if hasattr(mesh, "faces") and not mesh.faces.empty():
            nquad, ntri, nother = count_faces(mesh)
            if nother>0:
                warnings.warn("faces that are not triangle or quad detected.\nExporting as .mesh will result in a loss of data.")

            # export triangles
            if ntri>0:
                f.write("Triangles\n{}\n".format(ntri))
                for face in mesh.faces:
                    if len(face)==3:
                        f.write("{} {} {} 1\n".format(*(i+1 for i in face)))
                f.write("\n")

            # export quads
            if nquad>0:
                f.write("Quadrilaterals\n{}\n".format(nquad))
                for face in mesh.faces:
                    if len(face)==4:
                        f.write("{} {} {} {} 1\n".format(*(i+1 for i in face)))
                f.write("\n")
            
        if hasattr(mesh, "cells") and not mesh.cells.empty():
            nhex,ntet, nother = count_cells(mesh)
            if nother>0:
                warnings.warn("[Cells that are not tetrahedron or hexahedron detected.\nExporting as .mesh will result in a loss of data.")
            
            # export hexahedra
            if nhex>0:
                f.write("Hexahedra\n{}\n".format(nhex))
                for c in mesh.cells:
                    if len(c)==8:
                        f.write("{} {} {} {} {} {} {} {} 1\n".format(*(i+1 for i in c)))
                f.write("\n")
            
            # export tetrahedra
            if ntet>0:
                f.write("Tetrahedra\n{}\n".format(ntet))
                for c in mesh.cells:
                    if len(c)==4:
                        f.write("{} {} {} {} 1\n".format(*(i+1 for i in c)))
                f.write("\n")
    


import numpy as np
import scipy.sparse as sp

from ..geometry import *
from ..processing.connection import SurfaceConnectionFaces
from ..mesh.datatypes import *
from ..attributes.misc_faces import face_area

@allowed_mesh_types(SurfaceMesh)
def gradient(mesh: SurfaceMesh, conn: SurfaceConnectionFaces = None, as_complex: bool = True) -> sp.csc_matrix:
    """
    Computes the gradient operator, i.e. a |F| x |V| matrix G such that for any scalar function f defined over vertices of a surface mesh, Gf is its gradient inside faces.

    Gf maps to each face of the mesh either a vector of $\mathbb{R}^2$ or a complex number representing the gradient vector inside this face in local base.


    See [https://github.com/GCoiffier/mouette/blob/main/examples/gradient.py](https://github.com/GCoiffier/mouette/blob/main/examples/gradient.py)

    Args:
        mesh (SurfaceMesh): The input mesh
        conn (SurfaceConnectionFaces, optional): Connection objects specifying local bases of faces. Will be computed if not provided. Defaults to None.
        as_complex (bool, optional): whether the output is |F| complex values of 2|F| float values

    Raises:
        Exception: Fails if the mesh is not a triangulation
    
    Returns:
        scipy.sparse.csc_matrix: Gradient operator
    """
    if not mesh.is_triangular():
        raise Exception("Mesh is not a triangulation")
    if conn is None : conn = SurfaceConnectionFaces(mesh)
    area = face_area(mesh)

    N = len(mesh.vertices)
    M = len(mesh.faces)

    if as_complex:
        cols, rows, vals = np.zeros(3*M, dtype=int), np.zeros(3*M, dtype=int), np.zeros(3*M, dtype=complex)

        for iT,(A,B,C) in enumerate(mesh.faces):
            aT = 2*area[iT]
            xA,yA = conn.project(mesh.vertices[A], iT)
            xB,yB = conn.project(mesh.vertices[B], iT)
            xC,yC = conn.project(mesh.vertices[C], iT)

            rows[3*iT],   cols[3*iT],   vals[3*iT]   = iT, A, complex(yB-yC, xC-xB) / aT
            rows[3*iT+1], cols[3*iT+1], vals[3*iT+1] = iT, B, complex(yC-yA, xA-xC) / aT
            rows[3*iT+2], cols[3*iT+2], vals[3*iT+2] = iT, C, complex(yA-yB, xB-xA) / aT

    else:
        cols, rows, vals = np.zeros(6*M, dtype=int), np.zeros(6*M, dtype=int), np.zeros(6*M, dtype=float)
        for iT,(A,B,C) in enumerate(mesh.faces):
            aT = 2*area[iT]
            xA,yA = conn.project(mesh.vertices[A], iT)
            xB,yB = conn.project(mesh.vertices[B], iT)
            xC,yC = conn.project(mesh.vertices[C], iT)

            rows[6*iT],   cols[6*iT],   vals[6*iT]   = 2*iT,   A, (yB-yC) / aT
            rows[6*iT+1], cols[6*iT+1], vals[6*iT+1] = 2*iT+1, A, (xC-xB) / aT
            rows[6*iT+2], cols[6*iT+2], vals[6*iT+2] = 2*iT,   B, (yC-yA) / aT
            rows[6*iT+3], cols[6*iT+3], vals[6*iT+3] = 2*iT+1, B, (xA-xC) / aT
            rows[6*iT+4], cols[6*iT+4], vals[6*iT+4] = 2*iT,   C, (yA-yB) / aT
            rows[6*iT+5], cols[6*iT+5], vals[6*iT+5] = 2*iT+1, C, (xB-xA) / aT
        M *= 2
    return sp.csc_matrix((vals, (rows,cols)), shape=(M,N))


"""
Attributes:
    EYE (L4_SH): Coefficient of the identity frame in L4_SH
    LX (L4_Operator): Angular momentum operator for spherical harmonics representation, x-axis
    LY (L4_Operator): Angular momentum operator for spherical harmonics representation, y-axis
    LZ (L4_Operator): Angular momentum operator for spherical harmonics representation, z-axis

    Rxpi4 (L4_Operator): rotation of angle pi/4 around axis x
    Rypi4 (L4_Operator): rotation of angle pi/4 around axis y
    Rzpi4 (L4_Operator): rotation of angle pi/4 around axis z
    Rxzpi4 (L4_Operator): composition Rzpi4 * Rxpi4
    Rxpi2 (L4_Operator): rotation of angle pi/2 around axis x
    Rypi2 (L4_Operator): rotation of angle pi/2 around axis y
    Rzpi2 (L4_Operator): rotation of angle pi/2 around axis z

References:
    Algebraic Representations for Volumetric Frame Fields, Palmer et al. (Supplementary materials)
"""

import numpy as np
from scipy.linalg import expm
from scipy.spatial.transform import Rotation
from math import cos, sin, sqrt, tan, atan
from .vector import Vec
from .geometry import norm, dot
from .rotations import axis_rot_from_z

from typing import Annotated

L4_SH = Annotated[np.ndarray, 9]
"""
Type annotation for linear combinations of L4 Spherical Harmonics. 
Orthogonal frames are represented by 9 coefficients in the L4 band of the spherical harmonics basis. 
In practice, L4_SH is a size 9 numpy array.
"""

L4_Operator = Annotated[np.ndarray, (9,9)]
"""
Type annotation for linear operators acting on L4_SH (i.e. 9x9 matrices)
"""

LX = np.array([
    [0.,      0.,        0.,        0.,        0.,        0.,         0.,         -sqrt(2),   0.      ],
    [0.,      0.,        0.,        0.,        0.,        0.,         -sqrt(7/2), 0.,         -sqrt(2)],
    [0.,      0.,        0.,        0.,        0.,        -3/sqrt(2), 0.,         -sqrt(7/2), 0.      ],
    [0.,      0.,        0.,        0.,        -sqrt(10), 0.,         -3/sqrt(2), 0.,         0.      ],
    [0.,      0.,        0.,        sqrt(10),  0.,        0.,         0.,         0.,         0.      ],
    [0.,      0.,        3/sqrt(2), 0.,        0.,        0.,         0.,         0.,         0.      ],
    [0.,      sqrt(7/2), 0.,        3/sqrt(2), 0.,        0.,         0.,         0.,         0.      ],
    [sqrt(2), 0.,        sqrt(7/2), 0.,        0.,        0.,         0.,         0.,         0.      ],
    [0.,      sqrt(2),   0.,        0.,        0.,        0.,         0.,         0.,         0.      ],
])
# Angular momentum operator for spherical harmonics representation, x-axis


LY = np.array([
    [0.,       sqrt(2),    0.,         0.,        0.,       0.,        0.,         0.,         0.      ],
    [-sqrt(2), 0.,         sqrt(7/2),  0.,        0.,       0.,        0.,         0.,         0.      ],
    [0.,       -sqrt(7/2), 0.,         3/sqrt(2), 0.,       0.,        0.,         0.,         0.      ],
    [0.,       0.,         -3/sqrt(2), 0.,        0.,       0.,        0.,         0.,         0.      ],
    [0.,       0.,         0.,         0.,        0.,       -sqrt(10), 0.,         0.,         0.      ],
    [0.,       0.,         0.,         0.,        sqrt(10), 0.,        -3/sqrt(2), 0.,         0.      ],
    [0.,       0.,         0.,         0.,        0.,       3/sqrt(2), 0.,         -sqrt(7/2), 0.      ],
    [0.,       0.,         0.,         0.,        0.,       0.,        sqrt(7/2),  0.,         -sqrt(2)],
    [0.,       0.,         0.,         0.,        0.,       0.,        0.,         sqrt(2),    0.      ],
])
# Angular momentum operators for spherical harmonics representation, y-axis


LZ = np.array([
    [0.,  0.,  0.,  0., 0., 0., 0., 0., 4.],
    [0.,  0.,  0.,  0., 0., 0., 0., 3., 0.],
    [0.,  0.,  0.,  0., 0., 0., 2., 0., 0.],
    [0.,  0.,  0.,  0., 0., 1., 0., 0., 0.],
    [0.,  0.,  0.,  0., 0., 0., 0., 0., 0.],
    [0.,  0.,  0., -1., 0., 0., 0., 0., 0.],
    [0.,  0., -2.,  0., 0., 0., 0., 0., 0.],
    [0., -3.,  0.,  0., 0., 0., 0., 0., 0.],
    [-4., 0.,  0.,  0., 0., 0., 0., 0., 0.],
])
# Angular momentum operators for spherical harmonics representation, z-axis

Rxpi4  : L4_Operator = expm(np.pi/4 * LX)
Rypi4  : L4_Operator = expm(np.pi/4 * LY)
Rzpi4  : L4_Operator = expm(np.pi/4 * LZ)
Rxzpi4 : L4_Operator = Rzpi4 @ Rxpi4
Rxpi2  : L4_Operator = expm(np.pi/2 * LX)
Rypi2  : L4_Operator = expm(np.pi/2 * LY)
Rzpi4  : L4_Operator = expm(np.pi/2 * LZ)
# Hard-coded rotation matrices for common angles

def RZ(a : float) -> L4_Operator:
    """
    The L4 operator representing a rotation around z-axis of angle a

    Args:
        a (float): angle

    Returns:
        L4_Operator : 9x9 matrix
    """
    c = [cos(k*a) for k in range(5)]
    s = [sin(k*a) for k in range(5)]
    return np.array([
        [ c[4], 0.,   0.,   0.,   0., 0.,   0.,   0.,   s[4]],
        [ 0.,   c[3], 0.,   0.,   0., 0.,   0.,   s[3], 0.  ],
        [ 0.,   0.,   c[2], 0.,   0., 0.,   s[2], 0.,   0.  ],
        [ 0.,   0.,   0.,   c[1], 0., s[1], 0.,   0.,   0.  ],
        [ 0.,   0.,   0.,   0.,   1., 0.,   0.,   0.,   0.  ],
        [ 0.,   0.,   0.,  -s[1], 0., c[1], 0.,   0.,   0.  ],
        [ 0.,   0.,  -s[2], 0.,   0., 0.,   c[2], 0.,   0.  ],
        [ 0.,  -s[3], 0.,   0.,   0., 0.,   0.,   c[3], 0.  ],
        [-s[4], 0.,   0.,   0.,   0., 0.,   0.,   0.,   c[4]],
    ])

def RY(a : float) -> L4_Operator:
    """
    The L4 operator representing a rotation around y-axis of angle a

    Args:
        a (float): angle

    Returns:
        L4_Operator : 9x9 matrix
    """
    return Rxpi2.T @ RZ(a) @ Rxpi2

def RX(a : float) -> L4_Operator:
    """
    The L4 operator representing a rotation around x-axis of angle a

    Args:
        a (float): angle

    Returns:
        L4_Operator : 9x9 matrix
    """
    return Rypi2 @ RZ(a) @ Rypi2.T


EYE : L4_SH = np.array([0., 0., 0., 0., sqrt(7/12), 0., 0., 0., sqrt(5/12) ])
# Coefficient of the identity frame in L4_SH

def skew_matrix_from_rotvec(w : Vec) -> L4_Operator :
    """
    Given a rotation in angle-axis form, computes the 9x9 corresponding skew-symmetric matrix
    
    Args:
        w (Vec): a rotation axis vector where norm(v) represents the angle of rotation.

    Returns:
        L4_Operator: a 9x9 matrix
    """
    return w[0] * LX + w[1] * LY + w[2] * LZ

def rot_matrix_from_euler(w : Vec) -> L4_Operator :
    """
    Given three euler angles (XYZ), computes the 9x9 corresponding rotation matrix that performs the same rotation onto L4_SH coefficients

    Args:
        w (Vec): vector of size 3 representing three euler angles in convention XYZ.

    Returns:
        L4_Operator: a 9x9 matrix
    """
    return RX(w[0]) @ RY(w[1]) @ RZ(w[2])

def rot_matrix_from_rotvec(v : Vec) -> L4_Operator :
    """
    Given a rotation in angle-axis form, computes the 9x9 corresponding rotation matrix that performs the same rotation onto L4_SH coefficients,
    defined as the exponential of the corresponding skew-symmetric matrix.

    Args:
        v (Vec): a rotation axis vector where norm(v) represents the angle of rotation.

    Returns:
        L4_Operator: a 9x9 matrix
    """
    return expm(skew_matrix_from_rotvec(v))

def from_vec3(v : Vec) -> L4_SH:
    """
    Given a rotation in angle-axis form, computes the corresponding frame coefficients in L4_SH basis.

    Args:
        v (Vec): a rotation axis vector where norm(v) represents the angle of rotation.

    Returns:
        L4_SH: corresponding coefficients
    """
    if norm(v)<1e-8:  return np.array([0.,0.,0.,0.,1.,0.,0.,0.,0.])
    R = rot_matrix_from_rotvec(v)
    sh = R[:,4] # apply rotation to representation of (0,0,1) vector : 1 at 4th coeff and 0 otherwise -> dot product is fourth column of matrix
    return sh

def from_frame(frame : Rotation) -> L4_SH:
    """
    Given a rotation as a scipy.Rotation object, computes the corresponding frame coefficients in L4_SH basis.

    Args:
        frame (Rotation): a scipy rotation.

    Returns:
        L4_SH: corresponding coefficients
    """
    axis = Vec(frame.as_euler("XYZ"))
    if norm(axis)<1e-8: return EYE
    R = rot_matrix_from_euler(axis)
    sh = R.dot(EYE) # apply found rotation to the spherical harmonic reference
    return sh

def rotate_frame(sh : L4_SH, r : Rotation) -> L4_SH :
    """
    Applies a rotation to a frame decomposed in L4_SH basis.

    Args:
        sh (L4_SH): coefficients of the frame
        r (Rotation|Vec): rotation to be applied, either a scipy.Rotation object or a axis-angle representation.

    Returns:
        L4_SH : the rotated frame coefficients
    """
    if isinstance(r, Rotation):
        return Vec(rot_matrix_from_euler(r.as_rotvec()).dot(sh))
    return Vec(rot_matrix_from_euler(r).dot(sh))

def project_to_frame(sh : Vec, stop_threshold : float = 1e-8, max_iter=1000, nrml_cstr: Vec = None):
    """Given the coefficients in the spherical harmonic basis, finds the frames that correspond the most.
    Also recomputes spherical harmonics coefficients to a perfect match.
    Uses the Cayleigh transform to approximate the 9D rotation.

    Parameters:
        sh (Vec): the 9 coefficients representing the frame in spherical harmonics basis
        stop_threshold (float,optional) : Stoping criterion on the norm of the residual. Defaults to 1e-8.
        max_iter (int, optional): Maximum number of gradient steps. Defaults to 100.
        nrml_cstr (Vec) : a 3 dimensionnal normal direction, to constraint the projection to only be around an axis. Defaults to None.
    
    Returns:
        frame (scipy.spatial.transform.Rotation) : the obtained frame
        a (L4_SH) : the projected and updated spherical harmonics coefficients
    """
    if norm(sh)<1e-6: return Rotation.identity(), sh
    q = Vec.normalized(sh)
    
    if nrml_cstr is None:
        seeds = [EYE] + [R.dot(EYE) for R in (Rxpi4, Rypi4, Rzpi4, Rxzpi4)]
        pi4 = np.pi/4
        frames = [Rotation.identity(),
                Rotation.from_euler("x", pi4),
                Rotation.from_euler("y", pi4),
                Rotation.from_euler("z", pi4),
                Rotation.from_euler("xz", (pi4,pi4))]
        dots = [ np.dot(q, seed) for seed in seeds]
        seed_choice = np.argmax(dots)
        a = seeds[seed_choice]
        frame = frames[seed_choice]
        for _ in range(max_iter):
            s, b = (q+a)/2, (q-a)
            l = np.array([ LX.dot(s), LY.dot(s), LZ.dot(s)])
            if norm( l.dot(b))<stop_threshold: break
            x = np.linalg.lstsq(l.T, b, rcond=None)[0]
            x = Vec.normalized(x) * 2*atan(norm(x)/2)
            rot = Rotation.from_rotvec(x)
            frame = rot * frame
            x = rot.as_euler("XYZ")
            a = rot_matrix_from_euler(x).dot(a)
    else:
        ax = axis_rot_from_z(nrml_cstr)
        frames = [ R * Rotation.from_rotvec(ax) for R in 
                    (Rotation.identity(),
                     Rotation.from_rotvec(np.pi/4 * nrml_cstr))] 
        seeds = [from_frame(_frame) for _frame in frames]
        dots = [ np.dot(q, seed) for seed in seeds]
        seed_choice = np.argmax(dots)
        a = seeds[seed_choice]
        frame = frames[seed_choice]
        L = skew_matrix_from_rotvec(nrml_cstr)
        for _ in range(max_iter):
            s,b = L.dot(q+a), q-a
            if abs(s.dot(b))<stop_threshold: break
            alpha = dot(s,b)/dot(s,s)
            rot = Rotation.from_rotvec(2*atan(alpha)*nrml_cstr)
            frame = rot * frame
            x = rot.as_euler("XYZ")
            a = rot_matrix_from_euler(x).dot(a)
    return frame,a

def project_to_frame_grad(sh : Vec, lr : float = 1e-1, grad_threshold : float = 1e-4, dot_threshold = 1e-6, max_iter=1000):
    """Given the coefficients in the spherical harmonic basis, finds the frames that correspond the most. Computation is unfortunately not direct:
    starting from the reference frame, we perform a gradient descent on the l2 distance between spherical harmonics coefficients.

    Also recomputes spherical harmonics coefficients to a perfect match.
    Uses the linearization of the exponential to approximate 9D rotations.
    
    Warning:
        This algorithm is less precise and less efficient than `project_to_frame`. Use the latter instead.

    Parameters:
        sh (Vec): the 9 coefficients representing the frame in spherical harmonics basis
        lr (float, optional): Gradient descent learning rate. Defaults to 1.
        grad_threshold (float,optional) : Stoping criterion on the norm of the gradient. Defaults to 1e-4
        dot_threshold (float,optional) : Stoping criterion on the norm of the step. Defaults to 1e-6
        max_iter (int, optional): Maximum number of gradient steps. Defaults to 1000.
    
    Returns:
        frame (scipy.spatial.transform.Rotation) : the obtained frame
        a (Vec) : the projected and updated spherical harmonics coefficients
    """

    if norm(sh)<1e-6: return Rotation.identity(), sh
    Id = np.eye(9)
    q = Vec.normalized(sh)
    Rx, Ry, Rz = (Id + np.pi/4 * L for L in (LX, LY, LZ))
    qLX, qLY, qLZ = (np.dot(L.T, q) for L in (LX, LY, LZ))
    seeds = [EYE] + [R.dot(EYE) for R in (Rxpi4, Rypi4, Rzpi4, Rxzpi4)]
    pi4 = np.pi/4
    frames = [Rotation.identity(),
              Rotation.from_euler("x", pi4),
              Rotation.from_euler("y", pi4),
              Rotation.from_euler("z", pi4),
              Rotation.from_euler("xz", (pi4,pi4))]
    dots = [ np.dot(q, seed) for seed in seeds]
    seed_choice = np.argmax(dots)
    old_dot = dots[seed_choice]

    a = seeds[seed_choice]
    frame = frames[seed_choice]
    n_iter = 0
    while n_iter<max_iter:
        gx,gy,gz = ( np.dot(_q, a) for _q in (qLX,qLY,qLZ))
        grad = Vec( gx, gy, gz ) # gradient in point a
        if grad.norm()< grad_threshold: break

        step = lr*grad
        # technically, Rx, Ry and Rz should be exponentials of L1, L2, L3 with corresponding angles,
        # but we approximate at first order exp(X) = 1 + X + o(|X|)
        Rx = Id + step.x * LX
        Ry = Id + step.y * LY
        Rz = Id + step.z * LZ
        
        R = Rx @ Ry @ Rz
        a = R.dot(a)
        
        new_dot = np.dot(q,a)
        if abs(old_dot-new_dot)<dot_threshold: break
        old_dot = new_dot

        frame = Rotation.from_euler("XYZ", [step.x, step.y, step.z]) * frame
        n_iter += 1
    return frame,a

def orthogonality_energy(sh : np.ndarray) -> float:
    """Computes an energy that is 0 if and only if the given spherical harmonics has octahedral symetry (ie represents an orthogonal frame)
    The energy is quadratic in terms of the harmonic coefficients

    Parameters:
        sh (np.ndarray): a spherical harmonics (9 coefficients)

    Returns:
        (float): the computed energy
    """
    s1 = 28*(sh[0]*sh[0] + sh[8]*sh[8]) + 7*(sh[1]*sh[1] + sh[7]*sh[7]) - 8*(sh[2]*sh[2] + sh[6]*sh[6]) - 17*(sh[3]*sh[3] + sh[5]*sh[5]) - 20*sh[4]*sh[4]
    s1 *= sqrt(2)

    s2 = 28*(sh[0]*sh[1] + sh[7]*sh[8]) + 10*sqrt(7)*(sh[1]*sh[2] + sh[6]*sh[7]) + 18*(sh[2]*sh[3] + sh[5]*sh[6]) + 4*sqrt(5)*sh[4]*sh[5]
    s2 *= sqrt(3)

    s3 = 28*(sh[0]*sh[7] - sh[1]*sh[8]) + 10*sqrt(7)*(sh[1]*sh[6] - sh[2]*sh[7]) + 4*sqrt(5)*sh[4]*sh[3] + 18*(sh[2]*sh[5] - sh[3]*sh[6])
    s3 *= sqrt(3)

    s4 = 4*sqrt(7)*(sh[0]*sh[2] + sh[6]*sh[8]) + 6*sqrt(7)*(sh[1]*sh[3] + sh[5]*sh[7]) + 12*sqrt(5)*sh[4]*sh[6] + 10*(sh[3]*sh[3] - sh[5]*sh[5])
    s4 *= sqrt(6)

    s5 = 4*sqrt(7)*(sh[6]*sh[0] - sh[2]*sh[8]) + 6*sqrt(7)*(sh[1]*sh[5] - sh[3]*sh[7]) + 12*sqrt(5)*sh[2]*sh[4] - 20*sh[3]*sh[5]
    s5 *= sqrt(6)

    e = np.array((s1,s2,s3,s4,s5))
    return np.dot(e,e)


from osqp import OSQP
from scipy.sparse import identity

def get_osqp_lin_solver():
    try:
        inst = OSQP()
        inst.setup(P=identity(1, format="csc"), verbose=False, linsys_solver="mkl pardiso")
        return "mkl pardiso"
    except ValueError:
        return "qdldl"


import numpy as np

from ..mesh.mesh_attributes import Attribute, ArrayAttribute
from ..mesh.datatypes import *
from ..geometry.geometry import det_3x3

@allowed_mesh_types(VolumeMesh)
def cell_volume(mesh : VolumeMesh, name="volume", persistent=True , dense=True) -> Attribute:
    """
    Computes the volume of each cell.

    Parameters:
        mesh (VolumeMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "volume".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: one float per cell
    """
    if persistent :
        volume = mesh.cells.create_attribute(name, float, dense=dense)
    else:
        volume = ArrayAttribute(float, len(mesh.cells)) if dense else Attribute(float)
    if not mesh.is_tetrahedral():
        raise NotImplementedError
    for ic,(A,B,C,D) in enumerate(mesh.cells):
        pA,pB,pC,pD = (mesh.vertices[_v] for _v in (A,B,C,D))
        volume[ic]= abs(det_3x3(pA-pD,pB-pD,pC-pD))/6
    return volume

@allowed_mesh_types(VolumeMesh)
def cell_barycenter(mesh : VolumeMesh, name="barycenter", persistent=True, dense=True)-> Attribute:
    """
    Computes the barycenter point of each cell.

    Parameters:
        mesh (VolumeMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: 3D vector per cell
    """
    if persistent :
        bary = mesh.cells.create_attribute(name, float, 3, dense=dense)
    else:
        bary = ArrayAttribute(float, len(mesh.cells), 3) if dense else Attribute(float, 3)
    for iC,C in enumerate(mesh.cells):
        bary[iC] = sum(mesh.vertices[u] for u in C) / len(C)
    return bary

@allowed_mesh_types(VolumeMesh)
def cell_faces_on_boundary(mesh : VolumeMesh, name="boundary", persistent=True, dense=False)-> Attribute:
    """
    Integer flag on cells. For each cell, computes the number of its faces that lay on the boundary

    Parameters:
        mesh (VolumeMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "boundary".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to False
    
    Returns:
        Attribute: bool per cell
    """
    if persistent:
        bndf = mesh.cells.create_attribute(name, int, dense=dense)
    else:
        bndf = ArrayAttribute(int, len(mesh.cells)) if dense else Attribute(int)

    for iF in mesh.id_faces:
        if len(mesh.connectivity.face_to_cells(iF))==1 : 
            iC = mesh.connectivity.face_to_cells(iF)[0]
            bndf[iC] += 1
    return bndf


from scipy.spatial.transform import Rotation
from ..mesh.datatypes import *
from . import Vec
from .aabb import AABB
import numpy as np

def translate(mesh : Mesh, tr : Vec) -> Mesh:
    """
    Translates all vertices of the mesh by a fixed vector

    Parameters:
        mesh (Mesh): the input mesh. Can be any mesh data structure
        tr (Vec): the translation vector

    Returns:
        Mesh: the translated mesh
    """
    for i in mesh.id_vertices:
        mesh.vertices[i] += tr
    return mesh

def rotate(mesh : Mesh, rot : Rotation, orig : Vec = None) -> Mesh:
    """Rotates all vertices of the mesh by a given rotation around a given origin.

    Parameters:
        mesh (Mesh): the input mesh. Can be any mesh data structure
        rot (scipy.spatial.transform.Rotation|list): The rotation, given as an object or 3 Euler angles
        orig (Vec, optional): The origin of the rotation. If not provided, the function rotates around (0,0,0). Defaults to None.

    Returns:
        Mesh: _description_
    """
    if isinstance(rot, np.ndarray):
        assert rot.shape == (3,3)
        rot =  Rotation.from_matrix(rot)
    elif isinstance(rot, list) or isinstance(rot, tuple):
        assert len(rot)==3
        rot = Rotation.from_euler("xyz", rot)
    else:
        assert isinstance(rot, Rotation)

    if orig is None:
        orig = Vec.zeros(3)

    for i in mesh.id_vertices:
        mesh.vertices[i] = orig + rot.apply(mesh.vertices[i] - orig)
    return mesh

def scale(mesh : Mesh, factor : float, orig : Vec = None) -> Mesh:
    """Scales the mesh by a given factor around a given origin (fixed point)

    Parameters:
        mesh (Mesh): the input mesh. Can be any mesh data structure
        factor (float): scale factor. If its magnitude is < 1e-8, will print a warning.
        orig (Vec, optional): Fixed point of the scaling. If not provided, it is set at (0,0,0). Defaults to None.

    Returns:
        Mesh: the scaled input mesh.
    """
    if orig is None:
        orig = Vec.zeros(3)
    for i in mesh.id_vertices:
        mesh.vertices[i] = orig + factor*(mesh.vertices[i] - orig)
    return mesh

def scale_xyz(mesh : Mesh, fx : float = 1., fy : float = 1., fz : float = 1., orig : Vec = None) -> Mesh:
    """Scales the mesh independently along three axes.
    If any factor as a magnitude < 1e-8, will print a warning.

    Parameters:
        mesh (Mesh): the input mesh. Can be any mesh data structure
        fx (float, optional): Scale factor along x. Defaults to 1.
        fy (float, optional): Scale factor along y. Defaults to 1.
        fz (float, optional): Scale factor along z. Defaults to 1.
        orig (Vec, optional): Fixed point of the scaling. If not provided, it is set at (0,0,0). Defaults to None.

    Returns:
        Mesh: the scaled mesh.
    """
    if orig is None:
        orig = mesh.vertices[0]
    for i in mesh.id_vertices:
        Pi = mesh.vertices[i]
        mesh.vertices[i] = orig + Vec( fx*(Pi.x - orig.x), fy*(Pi.y - orig.y), fz *(Pi.z - orig.z))
    return mesh


def normalize(mesh : Mesh, center_at_zero : bool=True) -> Mesh:
    """Applies translation and global scaling so that the barycenter of the mesh becomes the origin and the mesh is contained in [-1,1]^3

    Args:
        mesh (Mesh): the input mesh
        center_at_zero (bool, optional): if set to True, will normalize in the cube [-1,1]^3. Otherwise, will normalize in [0,1]^3. Defaults to True.

    Returns:
        Mesh: the scaled and translated mesh
    """
    bounding = AABB.of_mesh(mesh)
    sc = 1/np.max(bounding.span)
    if center_at_zero:
        return scale(translate(mesh, -bounding.center), 2*sc)
    else:
        return scale(translate(mesh, -bounding.mini), sc)


def fit_into_unit_cube(mesh : Mesh) -> Mesh:
    """Applies translation and global scaling for the mesh to fit inside a cube [0;1]^3

    Args:
        mesh (Mesh): the input mesh. Can be any mesh data structure

    Note:
        Alias for `transform.normalize(mesh, center_at_zero=False)`

    Returns:
        Mesh: the scaled and translated mesh
    """
    return normalize(mesh, center_at_zero=False)


def translate_to_origin(mesh : Mesh) -> Mesh:
    """Translates all vertices of the mesh so that its barycenter is at origin (0., 0., 0.)

    Parameters:
        mesh (Mesh): the input mesh

    Returns:
        Mesh: the translated mesh
    """
    return translate(mesh, -sum(mesh.vertices)/len(mesh.vertices))


def flatten(mesh : Mesh, dim : int = None) -> Mesh:
    """
    Snaps to 0 one dimension to retrieve a flat 2D mesh.

    Parameters:
        mesh (Mesh): input mesh
        dim (int, optional): The dimension to flatten. If None, the function chooses the dimension which has the smallest variance.
        
    Returns:
        (Mesh): the modified input mesh
    """
    if dim is None:
        variances = []
        for i in range(3):
            variances.append(np.var([p[i] for p in mesh.vertices]))
        dim = np.argmin(variances)
    for i in mesh.id_vertices:
        mesh.vertices[i][dim] = 0.
    return mesh


import numpy as np
from abc import abstractmethod, ABC

from ..geometry import Vec
from aenum import MultiValueEnum

class _BaseAttribute(ABC):

    ########################################

    class InvalidTypeError(Exception):
        def __init__(self, typ):
            message = f"Type '{typ}' is not a valid type for a mesh attribute"
            super().__init__(message)

    class InvalidSizeError(Exception):
        def __init__(self, n,m):
            message = f"Attempted to set an attribute value of {n} elements on an attribute of element size {m}"
            super().__init__(message)

    class TypeNotMatchingError(Exception):
        def __init__(self, val, t1, t2):
            message = f"Tried to set value {val} of type '{t1}' inside an attribute of type '{t2}'"
            super().__init__(message)

    class DefaultValueTypeDoesNotMatchError(Exception):
        def __init__(self, dflt_val, typ):
            message = f"Default value '{dflt_val}' has type '{type(dflt_val)}'' while attribute has type '{typ}'"
            super().__init__(message)

    class OutOfBoundsError(Exception):
        def __init__(self,n, attr_size):
            message = f"Index {n} is out of bounds for attribute with size {attr_size}"
            super().__init__(message)

    ################ Typing ################

    SUPPORTED_TYPES = {
        bool, np.bool_,
        int, np.uint8, np.int32, np.int64, 
        float, np.float32, np.float64, 
        complex, 
        str
    }

    class Type(MultiValueEnum):
        Bool = bool, np.bool_
        Int = int, np.int32, np.uint8, np.int64
        Float = float, np.float32, np.float64
        Complex = complex
        String = str

        @classmethod
        def from_string(cls, txt : str):
            if txt in {"complex", "\"complex\""}:
                return cls.Complex
            if txt in {"double", "\"double\"", "float", "\"float\""}:
                return cls.Float
            if txt in {"index_t", "\"index_t\"", "int", "\"int\"", "\"signed_index_t\""}:
                return cls.Int
            if txt in {"bool", "\"bool\""}:
                return cls.Bool
            if txt in {"str", "string"}:
                return cls.String
            raise Exception(f"String '{txt}' corresponds to no attribute type")

        def to_string(self):
            if self.name.lower()=="float":
                return "double"
            return self.name.lower()

        def byte_size(self):
            return {
                "Bool" : 1,
                "Int" : 4,
                "Float" : 8
            }.get(self.name, None)

        @property
        def dtype(self):
            if self == Attribute.Type.String : 
                return "<U32" # String are limited to 32 characters max
            return self.value

        def default_value(self,n=1):
            if n==1:
                if self == Attribute.Type.Bool: return False
                if self == Attribute.Type.Int: return int(0)
                if self == Attribute.Type.Float: return float(0.)
                if self == Attribute.Type.Complex : return complex(0.,0.)
                if self == Attribute.Type.String : return ""
                raise Attribute.InvalidTypeError(self)
            return Vec([self.default_value(1)]*n)

    @staticmethod
    def _can_be_casted(ta : Type, tb : Type):
        """ Check for type compatibility and casting """
        if ta==tb : return True
        casts = {(Attribute.Type.Bool, Attribute.Type.Int),
                 (Attribute.Type.Bool, Attribute.Type.Float),
                 (Attribute.Type.Int, Attribute.Type.Float)}
        return (ta,tb) in casts

    ########################################

    def __init__(self, elem_type, elem_size:int=1, default_value=None):
        """
        __init__ method and the whole Attribute class are not supposed to be manipulated outside of the DataContainer class

        Parameters:
            elem_type : the type of the attribute (bool, int, float, complex, string)

            elem_size (int, optional): Number of elem_type objects to be stored per element. Defaults to 1.
            
            default_value (optional) : the default value of the attribute is n_elem is not specified. 
                If it is not specified either, it will correspond to the default value of the type provided in elem_type.
        """
        self.type : Attribute.Type = Attribute.Type(elem_type)
        self.elemsize : int = elem_size
        self._default_value = default_value
        self._check_default_value_type()
        self._data = None
    
    @property
    def default_value(self):
        if self._default_value is None:
            self._default_value = self.type.default_value(self.elemsize)
        return self._default_value

    def _check_default_value_type(self):
        if self._default_value is None : return
        if Attribute.Type(type(self._default_value)) != self.type:
            raise Attribute.DefaultValueTypeDoesNotMatchError(self._default_value, self.type)

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass
    
    @abstractmethod
    def __len__(self):
        pass

    def __repr__(self):
        return self._data.__repr__()

    def __str__(self):
        return str(self._data)

    @abstractmethod
    def __iter__(self):
        """
        Sparse attributes allow to iterate only over non-default elements
        """
        pass

    @abstractmethod
    def _expand(self, n : int):
        """Expands the storage capacity of the attributes. Adds `n` to self.n_elem
        Parameters:
            n (int) : number of new elements
        """
        pass

    @abstractmethod
    def as_array(self):
        pass

    def empty(self):
        """
        Check if an attribute is empty.
        For an attribute to be empty, its number of elements should not be fixed, and the dictionnary should be empty
        """
        return len(self)==0
    
    @abstractmethod
    def clear(self):
        """
        Empties the attribute. Frees the memory and ensures that all access return default value
        """
        pass


class Attribute(_BaseAttribute):
    def __init__(self, elem_type, elem_size:int=1, default_value=None):
        """
        __init__ method and the whole Attribute class are not supposed to be manipulated outside of the DataContainer class

        Parameters:
            elem_type : the type of the attribute (bool, int, float, complex, string)

            elem_size (int, optional): Number of elem_type objects to be stored per element. Defaults to 1.
            
            default_value (optional) : the default value of the attribute is n_elem is not specified. 
                If it is not specified either, it will correspond to the default value of the type provided in elem_type.
        """
        self.type : Attribute.Type = Attribute.Type(elem_type)
        self.elemsize : int = elem_size
        self._default_value = default_value
        self._check_default_value_type()
        self._data = dict()
    
    def __getitem__(self, key):
        if key in self._data:
            return self._data[key]
        return self.default_value

    def __setitem__(self, key, value):
        if self.elemsize>1:
            data = list(value)
            n = len(data)
            if not n==self.elemsize: 
                # provided number of element do not match elemsize
                raise Attribute.InvalidSizeError(n,self.elemsize)
        
            datatype = type(data[0])
            data_attr_type = Attribute.Type(datatype)  
            if not self._can_be_casted(data_attr_type, self.type):
                raise Attribute.TypeNotMatchingError(data, datatype, self.type)
            self._data[key] = Vec(data)
        
        else:
            datatype = type(value)
            data_attr_type = Attribute.Type(datatype)
            if not self._can_be_casted(data_attr_type, self.type):
                raise Attribute.TypeNotMatchingError(value, datatype, self.type)
            self._data[key] = value

    def _expand(self, n : int):
        """Expands the storage capacity of the attributes. Adds `n` to self.n_elem
        Parameters:
            n (int) : number of new elements
        """
        pass # Nothing to do when data is stored in a dict

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        """
        Sparse attributes allow to iterate only over non-default elements
        """
        return self._data.keys().__iter__()

    def as_array(self, container_size):
        out = np.full((container_size, self.elemsize), self.default_value, dtype= self.type.dtype)
        for i,x in self._data.items():
            out[i,:] = x
        return np.squeeze(out)

    def clear(self):
        """
        Empties the attribute. Frees the memory and ensures that all access return default value
        """
        self._data = dict()

class ArrayAttribute(_BaseAttribute):
    def __init__(self, elem_type, n_elem:int, elem_size:int=1, default_value=None):
        """
        __init__ method and the whole Attribute class are not supposed to be manipulated outside of the DataContainer class.
        An ArrayAttribute stores its values in a numpy array. This is a less flexible but safer approach than Attribute.

        Parameters:
            elem_type : the type of the attribute (bool, int, float, complex, string)

            n_elem (int): Total number of elements in the container. Should match the size of the DataContainer the attribute is stored in.

            elem_size (int, optional): Number of elem_type objects to be stored per element. Defaults to 1.
            
            dense (bool, optional): _description_. Defaults to False.

            default_value (optional) : the default value of the attribute is n_elem is not specified. 
                If it is not specified either, it will correspond to the default value of the type provided in elem_type.
        """
        self.type : Attribute.Type = Attribute.Type(elem_type)
        self.elemsize : int = elem_size
        self.n_elem:int = n_elem
        self._default_value = default_value
        self._check_default_value_type()
        self._data = np.full((n_elem, elem_size), self.default_value, dtype= self.type.dtype)
    
    def _check_out_of_bounds(self,key):
        if key<0 or key>self.n_elem:
            raise Attribute.OutOfBoundsError(key, self.n_elem)

    def __getitem__(self, key):
        self._check_out_of_bounds(key)
        return self._data[key,0] if self.elemsize==1 else self._data[key,:]

    def __setitem__(self, key, value):
        self._check_out_of_bounds(key)
        if self.elemsize>1:
            data = list(value)
            n = len(data)
            if not n==self.elemsize: 
                # provided number of element do not match elemsize
                raise Attribute.InvalidSizeError(n,self.elemsize)
        
            datatype = type(data[0])
            data_attr_type = Attribute.Type(datatype)  
            if not self._can_be_casted(data_attr_type, self.type):
                raise Attribute.TypeNotMatchingError(data, datatype, self.type)
            self._data[key] = Vec(data)
        
        else:
            datatype = type(value)
            data_attr_type = Attribute.Type(datatype)
            if not self._can_be_casted(data_attr_type, self.type):
                raise Attribute.TypeNotMatchingError(value, datatype, self.type)
            self._data[key] = value

    def _expand(self, n : int):
        """Expands the storage capacity of the attributes. Adds `n` to self.n_elem
        Parameters:
            n (int) : number of new elements
        """
        self._data = np.concatenate((self._data, np.full((n, self.elemsize), self.default_value, dtype= self.type.dtype)))
        self.n_elem += n

    def __len__(self):
        return self.n_elem

    def __iter__(self):
        """
        Sparse attributes allow to iterate only over non-default elements
        """
        return range(self.n_elem)

    def as_array(self, *args):
        return np.squeeze(self._data)
    
    def clear(self):
        """
        Empties the attribute. Frees the memory and ensures that all access return default value
        """
        self._data = np.full((self.n_elem, self.elemsize), self.default_value, dtype= self.type.dtype)